#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""The Store interface to a database.

This module contains the highest-level ORM interface in Storm.
"""

from copy import copy
from weakref import WeakValueDictionary
from operator import itemgetter

from storm.info import get_cls_info, get_obj_info, set_obj_info
from storm.variables import Variable, LazyValue
from storm.expr import (
    Expr, Select, Insert, Update, Delete, Column, Count, Max, Min,
    Avg, Sum, Eq, And, Asc, Desc, compile_python, compare_columns, SQLRaw,
    Union, Except, Intersect, Alias, SetExpr)
from storm.exceptions import (
    WrongStoreError, NotFlushedError, OrderLoopError, UnorderedError,
    NotOneError, FeatureError, CompileError, LostObjectError, ClassInfoError)
from storm.properties import PropertyColumn
from storm import Undef
from storm.cache import Cache
from storm.event import EventSystem


__all__ = [
    "AutoReload",
    "block_access",
    "EmptyResultSet",
    "Store",
    ]


PENDING_ADD = 1
PENDING_REMOVE = 2


class Store:
    """The Storm Store.

    This is the highest-level interface to a database. It manages
    transactions with L{commit} and L{rollback}, caching, high-level
    querying with L{find}, and more.

    Note that Store objects are not threadsafe. You should create one
    Store per thread in your application, passing them the same
    backend L{Database<storm.store.Database>} object.
    """

    _result_set_factory = None

    def __init__(self, database, cache=None):
        """
        @param database: The L{storm.database.Database} instance to use.
        @param cache: The cache to use.  Defaults to a L{Cache} instance.
        """
        self._database = database
        self._event = EventSystem(self)
        self._connection = database.connect(self._event)
        self._alive = WeakValueDictionary()
        self._dirty = {}
        self._order = {} # (info, info) = count
        if cache is None:
            self._cache = Cache()
        else:
            self._cache = cache
        self._implicit_flush_block_count = 0
        self._sequence = 0 # Advisory ordering.

    def get_database(self):
        """Return this Store's Database object."""
        return self._database

    @staticmethod
    def of(obj):
        """Get the Store that the object is associated with.

        If the given object has not yet been associated with a store,
        return None.
        """
        try:
            return get_obj_info(obj).get("store")
        except (AttributeError, ClassInfoError):
            return None

    def execute(self, statement, params=None, noresult=False):
        """Execute a basic query.

        This is just like L{storm.database.Connection.execute}, except
        that a flush is performed first.
        """
        if self._implicit_flush_block_count == 0:
            self.flush()
        return self._connection.execute(statement, params, noresult)

    def close(self):
        """Close the connection."""
        self._connection.close()

    def begin(self, xid):
        """Start a new two-phase transaction.

        @param xid: A L{Xid <storm.xid.Xid>} instance holding identification
            data for the new transaction.
        """
        self._connection.begin(xid)

    def prepare(self):
        """Prepare a two-phase transaction for the final commit.

        @note: It must be called inside a two-phase transaction started
            with L{begin}.
        """
        self._connection.prepare()

    def commit(self):
        """Commit all changes to the database.

        This invalidates the cache, so all live objects will have data
        reloaded next time they are touched.
        """
        self.flush()
        self.invalidate()
        self._connection.commit()

    def rollback(self):
        """Roll back all outstanding changes, reverting to database state."""
        for obj_info in self._dirty:
            pending = obj_info.pop("pending", None)
            if pending is PENDING_ADD:
                # Object never got in the cache, so being "in the store"
                # has no actual meaning for it.
                del obj_info["store"]
            elif pending is PENDING_REMOVE:
                # Object never got removed, so it's still in the cache,
                # and thus should continue to resolve from now on.
                self._enable_lazy_resolving(obj_info)
        self._dirty.clear()
        self.invalidate()
        self._connection.rollback()

    def get(self, cls, key):
        """Get object of type cls with the given primary key from the database.

        If the object is alive the database won't be touched.

        @param cls: Class of the object to be retrieved.
        @param key: Primary key of object. May be a tuple for composed keys.

        @return: The object found with the given primary key, or None
            if no object is found.
        """

        if self._implicit_flush_block_count == 0:
            self.flush()

        if type(key) != tuple:
            key = (key,)

        cls_info = get_cls_info(cls)

        assert len(key) == len(cls_info.primary_key)

        primary_vars = []
        for column, variable in zip(cls_info.primary_key, key):
            if not isinstance(variable, Variable):
                variable = column.variable_factory(value=variable)
            primary_vars.append(variable)

        primary_values = tuple(var.get(to_db=True) for var in primary_vars)
        obj_info = self._alive.get((cls_info.cls, primary_values))
        if obj_info is not None and not obj_info.get("invalidated"):
            return self._get_object(obj_info)

        where = compare_columns(cls_info.primary_key, primary_vars)

        select = Select(cls_info.columns, where,
                        default_tables=cls_info.table, limit=1)

        result = self._connection.execute(select)
        values = result.get_one()
        if values is None:
            return None
        return self._load_object(cls_info, result, values)

    def find(self, cls_spec, *args, **kwargs):
        """Perform a query.

        Some examples::

            store.find(Person, Person.name == u"Joe") --> all Persons named Joe
            store.find(Person, name=u"Joe") --> same
            store.find((Company, Person), Person.company_id == Company.id) -->
                iterator of tuples of Company and Person instances which are
                associated via the company_id -> Company relation.

        @param cls_spec: The class or tuple of classes whose
            associated tables will be queried.
        @param args: Instances of L{Expr}.
        @param kwargs: Mapping of simple column names to values or
            expressions to query for.

        @return: A L{ResultSet} of instances C{cls_spec}. If C{cls_spec}
            was a tuple, then an iterator of tuples of such instances.
        """
        if self._implicit_flush_block_count == 0:
            self.flush()
        find_spec = FindSpec(cls_spec)
        where = get_where_for_args(args, kwargs, find_spec.default_cls)
        return self._result_set_factory(self, find_spec, where)

    def using(self, *tables):
        """Specify tables to use explicitly.

        The L{find} method generally does a good job at figuring out
        the tables to query by itself, but in some cases it's useful
        to specify them explicitly.

        This is most often necessary when an explicit SQL join is
        required. An example follows::

            join = LeftJoin(Person, Person.id == Company.person_id)
            print(list(store.using(Company, join).find((Company, Person))))

        The previous code snippet will produce an SQL statement
        somewhat similar to this, depending on your backend::

            SELECT company.id, employee.company_id, employee.id
            FROM company
            LEFT JOIN employee ON employee.company_id = company.id;

        @return: A L{TableSet}, which has a C{find} method similar to
            L{Store.find}.
        """
        return self._table_set(self, tables)

    def add(self, obj):
        """Add the given object to the store.

        The object will be inserted into the database if it has not
        yet been added.

        The C{added} event will be fired on the object info's event system.
        """
        self._event.emit("register-transaction")
        obj_info = get_obj_info(obj)

        store = obj_info.get("store")
        if store is not None and store is not self:
            raise WrongStoreError("%s is part of another store" % repr(obj))

        pending = obj_info.get("pending")

        if pending is PENDING_ADD:
            pass
        elif pending is PENDING_REMOVE:
            del obj_info["pending"]
            self._enable_lazy_resolving(obj_info)
            # obj_info.event.emit("added")
        elif store is None:
            obj_info["store"] = self
            obj_info["pending"] = PENDING_ADD
            self._set_dirty(obj_info)
            self._enable_lazy_resolving(obj_info)
            obj_info.event.emit("added")

        return obj

    def remove(self, obj):
        """Remove the given object from the store.

        The associated row will be deleted from the database.
        """
        self._event.emit("register-transaction")
        obj_info = get_obj_info(obj)

        if obj_info.get("store") is not self:
            raise WrongStoreError("%s is not in this store" % repr(obj))

        pending = obj_info.get("pending")

        if pending is PENDING_REMOVE:
            pass
        elif pending is PENDING_ADD:
            del obj_info["store"]
            del obj_info["pending"]
            self._set_clean(obj_info)
            self._disable_lazy_resolving(obj_info)
            obj_info.event.emit("removed")
        else:
            obj_info["pending"] = PENDING_REMOVE
            self._set_dirty(obj_info)
            self._disable_lazy_resolving(obj_info)
            obj_info.event.emit("removed")

    def reload(self, obj):
        """Reload the given object.

        The object will immediately have all of its data reset from
        the database. Any pending changes will be thrown away.
        """
        obj_info = get_obj_info(obj)
        cls_info = obj_info.cls_info
        if obj_info.get("store") is not self:
            raise WrongStoreError("%s is not in this store" % repr(obj))
        if "primary_vars" not in obj_info:
            raise NotFlushedError("Can't reload an object if it was "
                                  "never flushed")
        where = compare_columns(cls_info.primary_key, obj_info["primary_vars"])
        select = Select(cls_info.columns, where,
                        default_tables=cls_info.table, limit=1)
        result = self._connection.execute(select)
        values = result.get_one()
        self._set_values(obj_info, cls_info.columns, result, values,
                         replace_unknown_lazy=True)
        self._set_clean(obj_info)

    def autoreload(self, obj=None):
        """Set an object or all objects to be reloaded automatically on access.

        When a database-backed attribute of one of the objects is
        accessed, the object will be reloaded entirely from the database.

        @param obj: If passed, only mark the given object for
            autoreload. Otherwise, all cached objects will be marked for
            autoreload.
        """
        self._mark_autoreload(obj, False)

    def invalidate(self, obj=None):
        """Set an object or all objects to be invalidated.

        This prevents Storm from returning the cached object without
        first verifying that the object is still available in the
        database.

        This should almost never be called by application code; it is
        only necessary if it is possible that an object has
        disappeared through some mechanism that Storm was unable to
        detect, like direct SQL statements within the current
        transaction that bypassed the ORM layer. The Store
        automatically invalidates all cached objects on transaction
        boundaries.
        """
        if obj is None:
            self._cache.clear()
        else:
            self._cache.remove(get_obj_info(obj))
        self._mark_autoreload(obj, True)

    def reset(self):
        """Reset this store, causing all future queries to return new objects.

        Beware this method: it breaks the assumption that there will never be
        two objects in memory which represent the same database object.

        This is useful if you've got in-memory changes to an object that you
        want to "throw out"; next time they're fetched the objects will be
        recreated, so in-memory modifications will not be in effect for future
        queries.
        """
        for obj_info in self._iter_alive():
            if "store" in obj_info:
                del obj_info["store"]
        self._alive.clear()
        self._dirty.clear()
        self._cache.clear()
        # The following line is untested, but then, I can't really find a way
        # to test it without whitebox.
        self._order.clear()


    def _mark_autoreload(self, obj=None, invalidate=False):
        if obj is None:
            obj_infos = self._iter_alive()
        else:
            obj_infos = (get_obj_info(obj),)
        for obj_info in obj_infos:
            cls_info = obj_info.cls_info
            for column in cls_info.columns:
                if id(column) not in cls_info.primary_key_idx:
                    obj_info.variables[column].set(AutoReload)
            if invalidate:
                # Marking an object with 'invalidated' means that we're
                # not sure if the object is actually in the database
                # anymore, so before the object is returned from the cache
                # (e.g. by a get()), the database should be queried to see
                # if the object's still there.
                obj_info["invalidated"] = True
        # We want to make sure we've marked all objects as invalidated and set
        # up their autoreloads before calling the invalidated hook on *any* of
        # them, because an invalidated hook might use other objects and we want
        # to prevent invalidation ordering issues.
        if invalidate:
            for obj_info in obj_infos:
                self._run_hook(obj_info, "__storm_invalidated__")

    def add_flush_order(self, before, after):
        """Explicitly specify the order of flushing two objects.

        When the next database flush occurs, the order of data
        modification statements will be ensured.

        @param before: The object to flush first.
        @param after: The object to flush after C{before}.
        """
        pair = (get_obj_info(before), get_obj_info(after))
        try:
            self._order[pair] += 1
        except KeyError:
            self._order[pair] = 1

    def remove_flush_order(self, before, after):
        """Cancel an explicit flush order specified with L{add_flush_order}.

        @param before: The C{before} object previously specified in a
            call to L{add_flush_order}.
        @param after: The C{after} object previously specified in a
            call to L{add_flush_order}.
        """
        pair = (get_obj_info(before), get_obj_info(after))
        self._order[pair] -= 1

    def flush(self):
        """Flush all dirty objects in cache to database.

        This method will first call the __storm_pre_flush__ hook of all dirty
        objects.  If more objects become dirty as a result of executing code
        in the hooks, the hook is also called on them.  The hook is only
        called once for each object.

        It will then flush each dirty object to the database, that is,
        execute the SQL code to insert/delete/update them.  After each
        object is flushed, the hook __storm_flushed__ is called on it,
        and if changes are made to the object it will get back to the
        dirty list, and be flushed again.

        Note that Storm will flush objects for you automatically, so you'll
        only need to call this method explicitly in very rare cases where
        normal flushing times are insufficient, such as when you want to
        make sure a database trigger gets run at a particular time.
        """
        self._event.emit("flush")

        # The _dirty list may change under us while we're running
        # the flush hooks, so we cannot just simply loop over it
        # once.  To prevent infinite looping we keep track of which
        # objects we've called the hook for using a `flushing` dict.
        flushing = {}
        while self._dirty:
            (obj_info, obj) = self._dirty.popitem()
            if obj_info not in flushing:
                flushing[obj_info] = obj
                self._run_hook(obj_info, "__storm_pre_flush__")
        self._dirty = flushing

        predecessors = {}
        for (before_info, after_info), n in self._order.items():
            if n > 0:
                before_set = predecessors.get(after_info)
                if before_set is None:
                    predecessors[after_info] = {before_info}
                else:
                    before_set.add(before_info)

        key_func = itemgetter("sequence")

        # The external loop is important because items can get into the dirty
        # state while we're flushing objects, ...
        while self._dirty:
            # ... but we don't have to resort everytime an object is flushed,
            # so we have an internal loop too.  If no objects become dirty
            # during flush, this will clean self._dirty and the external loop
            # will exit too.
            sorted_dirty = sorted(self._dirty, key=key_func)
            while sorted_dirty:
                for i, obj_info in enumerate(sorted_dirty):
                    for before_info in predecessors.get(obj_info, ()):
                        if before_info in self._dirty:
                            break # A predecessor is still dirty.
                    else:
                        break # Found an item without dirty predecessors.
                else:
                    raise OrderLoopError("Can't flush due to ordering loop")
                del sorted_dirty[i]
                self._dirty.pop(obj_info, None)
                self._flush_one(obj_info)

        self._order.clear()

        # That's not stricly necessary, but prevents getting into bigints.
        self._sequence = 0

    def _flush_one(self, obj_info):
        cls_info = obj_info.cls_info

        pending = obj_info.pop("pending", None)

        if pending is PENDING_REMOVE:
            expr = Delete(compare_columns(cls_info.primary_key,
                                          obj_info["primary_vars"]),
                          cls_info.table)
            self._connection.execute(expr, noresult=True)

            # We're sure the cache is valid at this point.
            obj_info.pop("invalidated", None)

            self._disable_change_notification(obj_info)
            self._remove_from_alive(obj_info)
            del obj_info["store"]

        elif pending is PENDING_ADD:

            # Give a chance to the backend to process primary variables.
            self._connection.preset_primary_key(cls_info.primary_key,
                                                obj_info.primary_vars)

            changes = self._get_changes_map(obj_info, True)

            expr = Insert(changes, cls_info.table,
                          primary_columns=cls_info.primary_key,
                          primary_variables=obj_info.primary_vars)

            result = self._connection.execute(expr)

            # We're sure the cache is valid at this point. We just added
            # the object.
            obj_info.pop("invalidated", None)

            self._fill_missing_values(obj_info, obj_info.primary_vars, result)

            self._enable_change_notification(obj_info)
            self._add_to_alive(obj_info)
        else:
            cached_primary_vars = obj_info["primary_vars"]

            changes = self._get_changes_map(obj_info)

            if changes:
                expr = Update(changes,
                              compare_columns(cls_info.primary_key,
                                              cached_primary_vars),
                              cls_info.table)
                self._connection.execute(expr, noresult=True)

                self._fill_missing_values(obj_info, obj_info.primary_vars)

                self._add_to_alive(obj_info)

        self._run_hook(obj_info, "__storm_flushed__")

        obj_info.event.emit("flushed")

    def block_implicit_flushes(self):
        """Block implicit flushes from operations like execute()."""
        self._implicit_flush_block_count += 1

    def unblock_implicit_flushes(self):
        """Unblock implicit flushes from operations like execute()."""
        assert self._implicit_flush_block_count > 0
        self._implicit_flush_block_count -= 1

    def block_access(self):
        """Block access to the underlying database connection."""
        self._connection.block_access()

    def unblock_access(self):
        """Unblock access to the underlying database connection."""
        self._connection.unblock_access()

    def _get_changes_map(self, obj_info, adding=False):
        """Return a {column: variable} dictionary suitable for inserts/updates.

        @param obj_info: ObjectInfo to inspect for changes.
        @param adding: If true, any defined variables will be considered
                       a change and included in the returned map.
        """
        cls_info = obj_info.cls_info
        changes = {}
        select_variables = []
        for column in cls_info.columns:
            variable = obj_info.variables[column]
            if adding or variable.has_changed():
                if variable.is_defined():
                    changes[column] = variable
                else:
                    lazy_value = variable.get_lazy()
                    if isinstance(lazy_value, Expr):
                        if id(column) in cls_info.primary_key_idx:
                            select_variables.append(variable) # See below.
                            changes[column] = variable
                        else:
                            changes[column] = lazy_value

        # If we have any expressions in the primary variables, we
        # have to resolve them now so that we have the identity of
        # the inserted object available later.
        if select_variables:
            resolve_expr = Select([variable.get_lazy()
                                   for variable in select_variables])
            result = self._connection.execute(resolve_expr)
            for variable, value in zip(select_variables, result.get_one()):
                result.set_variable(variable, value)

        return changes

    def _fill_missing_values(self, obj_info, primary_vars, result=None):
        """Fill missing values in variables of the given obj_info.

        This method will verify which values are unset in obj_info,
        and set them to L{AutoReload}, or if it's part of the primary
        key, query the database for the actual values.

        @param obj_info: ObjectInfo to have its values filled.
        @param primary_vars: Variables composing the primary key with
            up-to-date values (cached variables may be out-of-date when
            this method is called).
        @param result: If some value in the set of primary variables
            isn't defined, it must be retrieved from the database
            using database-dependent logic, which is provided by the
            backend in the result of the query which inserted the object.
        """
        cls_info = obj_info.cls_info

        cached_primary_vars = obj_info.get("primary_vars")
        primary_key_idx = cls_info.primary_key_idx
        missing_columns = []
        for column in cls_info.columns:
            variable = obj_info.variables[column]
            if not variable.is_defined():
                idx = primary_key_idx.get(id(column))
                if idx is not None:
                    if (cached_primary_vars is not None
                        and variable.get_lazy() is AutoReload):
                        # For auto-reloading a primary key, just
                        # get the value out of the cache.
                        variable.set(cached_primary_vars[idx].get())
                    else:
                        missing_columns.append(column)
                else:
                    # Any lazy values are overwritten here.  This value
                    # must have just been sent to the database, so this
                    # was already set there.
                    variable.set(AutoReload)
            else:
                variable.checkpoint()

        if missing_columns:
            where = result.get_insert_identity(cls_info.primary_key,
                                               primary_vars)
            result = self._connection.execute(Select(missing_columns, where))
            self._set_values(obj_info, missing_columns,
                             result, result.get_one())

    def _validate_alive(self, obj_info):
        """Perform cache validation for the given obj_info."""
        where = compare_columns(obj_info.cls_info.primary_key,
                                obj_info["primary_vars"])
        result = self._connection.execute(Select(SQLRaw("1"), where))
        if not result.get_one():
            raise LostObjectError("Object is not in the database anymore")
        obj_info.pop("invalidated", None)

    def _load_object(self, cls_info, result, values):
        # _set_values() need the cls_info columns for the class of the
        # actual object, not from a possible wrapper (e.g. an alias).
        cls = cls_info.cls
        cls_info = get_cls_info(cls)

        # Prepare cache key.
        primary_vars = []
        columns = cls_info.columns

        for value in values:
            if value is not None:
                break
        else:
            # We've got a row full of NULLs, so consider that the object
            # wasn't found.  This is useful for joins, where non-existent
            # rows are represented like that.
            return None

        for i in cls_info.primary_key_pos:
            value = values[i]
            variable = columns[i].variable_factory(value=value, from_db=True)
            primary_vars.append(variable)

        # Lookup cache.
        primary_values = tuple(var.get(to_db=True) for var in primary_vars)
        obj_info = self._alive.get((cls, primary_values))

        if obj_info is not None:
            # Found object in cache, and it must be valid since the
            # primary key was extracted from result values.
            obj_info.pop("invalidated", None)

            # Take that chance and fill up any undefined variables
            # with fresh data, since we got it anyway.
            self._set_values(obj_info, cls_info.columns, result,
                             values, keep_defined=True)

            # We're not sure if the obj is still in memory at this
            # point.  This will rebuild it if needed.
            obj = self._get_object(obj_info)
        else:
            # Nothing found in the cache. Build everything from the ground.
            obj = cls.__new__(cls)

            obj_info = get_obj_info(obj)
            obj_info["store"] = self

            self._set_values(obj_info, cls_info.columns, result, values,
                             replace_unknown_lazy=True)

            self._add_to_alive(obj_info)
            self._enable_change_notification(obj_info)
            self._enable_lazy_resolving(obj_info)

            self._run_hook(obj_info, "__storm_loaded__")

        return obj

    def _get_object(self, obj_info):
        """Return object for obj_info, rebuilding it if it's dead."""
        obj = obj_info.get_obj()
        if obj is None:
            cls = obj_info.cls_info.cls
            obj = cls.__new__(cls)
            obj_info.set_obj(obj)
            set_obj_info(obj, obj_info)
            # Re-enable change notification, as it may have been implicitely
            # disabled when the previous object has been collected
            self._enable_change_notification(obj_info)
            self._run_hook(obj_info, "__storm_loaded__")
        # Renew the cache.
        self._cache.add(obj_info)
        return obj

    @staticmethod
    def _run_hook(obj_info, hook_name):
        func = getattr(obj_info.get_obj(), hook_name, None)
        if func is not None:
            func()

    def _set_values(self, obj_info, columns, result, values,
                    keep_defined=False, replace_unknown_lazy=False):
        if values is None:
            raise LostObjectError("Can't obtain values from the database "
                                  "(object got removed?)")
        obj_info.pop("invalidated", None)
        for column, value in zip(columns, values):
            variable = obj_info.variables[column]
            lazy_value = variable.get_lazy()
            is_unknown_lazy = not (lazy_value is None or
                                   lazy_value is AutoReload)
            if keep_defined:
                if variable.is_defined() or is_unknown_lazy:
                    continue
            elif is_unknown_lazy and not replace_unknown_lazy:
                # This should *never* happen, because whenever we get
                # to this point it should be after a flush() which
                # updated the database with lazy values and then replaced
                # them by AutoReload.  Letting this go through means
                # we're blindly discarding an unknown lazy value and
                # replacing it by the value from the database.
                raise RuntimeError("Unexpected situation. "
                                   "Please contact the developers.")
            if value is None:
                variable.set(value, from_db=True)
            else:
                result.set_variable(variable, value)

            variable.checkpoint()


    def _is_dirty(self, obj_info):
        return obj_info in self._dirty

    def _set_dirty(self, obj_info):
        if obj_info not in self._dirty:
            self._dirty[obj_info] = obj_info.get_obj()
            obj_info["sequence"] = self._sequence = self._sequence + 1

    def _set_clean(self, obj_info):
        self._dirty.pop(obj_info, None)

    def _iter_dirty(self):
        return self._dirty


    def _add_to_alive(self, obj_info):
        """Add an object to the set of known in-memory objects.

        When an object is added to the set of known in-memory objects,
        the key is built from a copy of the current variables that are
        part of the primary key.  This means that, when an object is
        retrieved from the database, these values may be used to get
        the cached object which is already in memory, even if it
        requested the primary key value to be changed.  For that reason,
        when changes to the primary key are flushed, the alive object
        key should also be updated to reflect these changes.

        In addition to tracking objects alive in memory, we have a strong
        reference cache which keeps a fixed number of last-used objects
        in-memory, to prevent further database access for recently fetched
        objects.
        """
        cls_info = obj_info.cls_info
        old_primary_vars = obj_info.get("primary_vars")
        if old_primary_vars is not None:
            old_primary_values = tuple(
                var.get(to_db=True) for var in old_primary_vars)
            self._alive.pop((cls_info.cls, old_primary_values), None)
        new_primary_vars = tuple(variable.copy()
                                 for variable in obj_info.primary_vars)
        new_primary_values = tuple(
            var.get(to_db=True) for var in new_primary_vars)
        self._alive[cls_info.cls, new_primary_values] = obj_info
        obj_info["primary_vars"] = new_primary_vars
        self._cache.add(obj_info)

    def _remove_from_alive(self, obj_info):
        """Remove an object from the cache.

        This method is only called for objects that were explicitly
        deleted and flushed.  Objects that are unused will get removed
        from the cache dictionary automatically by their weakref callbacks.
        """
        primary_vars = obj_info.get("primary_vars")
        if primary_vars is not None:
            self._cache.remove(obj_info)
            primary_values = tuple(var.get(to_db=True) for var in primary_vars)
            del self._alive[obj_info.cls_info.cls, primary_values]
            del obj_info["primary_vars"]

    def _iter_alive(self):
        return list(self._alive.values())

    def _enable_change_notification(self, obj_info):
        obj_info.event.emit("start-tracking-changes", self._event)
        obj_info.event.hook("changed", self._variable_changed)

    def _disable_change_notification(self, obj_info):
        obj_info.event.unhook("changed", self._variable_changed)
        obj_info.event.emit("stop-tracking-changes", self._event)

    def _variable_changed(self, obj_info, variable,
                          old_value, new_value, fromdb):
        # The fromdb check makes sure that values coming from the
        # database don't mark the object as dirty again.
        # XXX The fromdb check is untested. How to test it?
        if not fromdb:
            if new_value is not Undef and new_value is not AutoReload:
                if obj_info.get("invalidated"):
                    # This might be a previously alive object being
                    # updated.  Let's validate it now to improve debugging.
                    # This will raise LostObjectError if the object is gone.
                    self._validate_alive(obj_info)
                self._set_dirty(obj_info)


    def _enable_lazy_resolving(self, obj_info):
        obj_info.event.hook("resolve-lazy-value", self._resolve_lazy_value)

    def _disable_lazy_resolving(self, obj_info):
        obj_info.event.unhook("resolve-lazy-value", self._resolve_lazy_value)

    def _resolve_lazy_value(self, obj_info, variable, lazy_value):
        """Resolve a variable set to a lazy value when it's touched.

        This method is hooked into the obj_info to resolve variables
        set to lazy values when they're accessed.  It will first flush
        the store, and then set all variables set to L{AutoReload} to
        their database values.
        """
        if lazy_value is not AutoReload and not isinstance(lazy_value, Expr):
            # It's not something we handle.
            return

        # XXX This will do it for now, but it should really flush
        #     just this single object and ones that it depends on.
        #     _flush_one() doesn't consider dependencies, so it may
        #     not be used directly.  Maybe allow flush(obj)?
        if self._implicit_flush_block_count == 0:
            self.flush()

        autoreload_columns = []
        for column in obj_info.cls_info.columns:
            if obj_info.variables[column].get_lazy() is AutoReload:
                autoreload_columns.append(column)

        if autoreload_columns:
            where = compare_columns(obj_info.cls_info.primary_key,
                                    obj_info["primary_vars"])
            result = self._connection.execute(
                Select(autoreload_columns, where))
            self._set_values(obj_info, autoreload_columns,
                             result, result.get_one())


class ResultSet:
    """The representation of the results of a query.

    Note that having an instance of this class does not indicate that
    a database query has necessarily been made. Database queries are
    put off until absolutely necessary.

    Generally these should not be constructed directly, but instead
    retrieved from calls to L{Store.find}.
    """

    def __init__(self, store, find_spec,
                 where=Undef, tables=Undef, select=Undef):
        self._store = store
        self._find_spec = find_spec
        self._where = where
        self._tables = tables
        self._select = select
        self._order_by = find_spec.default_order
        self._offset = Undef
        self._limit = Undef
        self._distinct = False
        self._group_by = Undef
        self._having = Undef

    def copy(self):
        """Return a copy of this ResultSet object, with the same configuration.
        """
        result_set = object.__new__(self.__class__)
        result_set.__dict__.update(self.__dict__)
        if self._select is not Undef:
            # This expression must be copied because we may have to change it
            # in-place inside _get_select().
            result_set._select = copy(self._select)
        return result_set

    def config(self, distinct=None, offset=None, limit=None):
        """Configure this result object in-place. All parameters are optional.

        @param distinct: If True, enables usage of the DISTINCT keyword in
            the query. If a tuple or list of columns, inserts a DISTINCT ON
            (only supported by PostgreSQL).
        @param offset: Offset where results will start to be retrieved
            from the result set.
        @param limit: Limit the number of objects retrieved from the
            result set.

        @return: self (not a copy).
        """

        if distinct is not None:
            self._distinct = distinct
        if offset is not None:
            self._offset = offset
        if limit is not None:
            self._limit = limit
        return self

    def _get_select(self):
        if self._select is not Undef:
            if self._order_by is not Undef:
                self._select.order_by = self._order_by
            if self._limit is not Undef: # XXX UNTESTED!
                self._select.limit = self._limit
            if self._offset is not Undef: # XXX UNTESTED!
                self._select.offset = self._offset
            return self._select
        columns, default_tables = self._find_spec.get_columns_and_tables()
        return Select(columns, self._where, self._tables, default_tables,
                      self._order_by, offset=self._offset, limit=self._limit,
                      distinct=self._distinct, group_by=self._group_by,
                      having=self._having)

    def _load_objects(self, result, values):
        return self._find_spec.load_objects(self._store, result, values)

    def __iter__(self):
        """Iterate the results of the query.
        """
        result = self._store._connection.execute(self._get_select())
        for values in result:
            yield self._load_objects(result, values)

    def __getitem__(self, index):
        """Get an individual item by offset, or a range of items by slice.

        @return: The matching object or, if a slice is used, a new
            L{ResultSet} will be returned appropriately modified with
            C{OFFSET} and C{LIMIT} clauses.
        """
        if isinstance(index, int):
            if index == 0:
                result_set = self
            else:
                if self._offset is not Undef:
                    index += self._offset
                result_set = self.copy()
                result_set.config(offset=index, limit=1)
            obj = result_set._any()
            if obj is None:
                raise IndexError("Index out of range")
            return obj

        if not isinstance(index, slice):
            raise IndexError("Can't index ResultSets with %r" % (index,))
        if index.step is not None:
            raise IndexError("Stepped slices not yet supported: %r"
                             % (index.step,))

        offset = self._offset
        limit = self._limit

        if index.start is not None:
            if offset is Undef:
                offset = index.start
            else:
                offset += index.start
            if limit is not Undef:
                limit = max(0, limit - index.start)

        if index.stop is not None:
            if index.start is None:
                new_limit = index.stop
            else:
                new_limit = index.stop - index.start
            if limit is Undef or limit > new_limit:
                limit = new_limit

        return self.copy().config(offset=offset, limit=limit)

    def __contains__(self, item):
        """Check if an item is contained within the result set."""
        columns, values = self._find_spec.get_columns_and_values_for_item(item)

        if self._select is Undef and self._group_by is Undef:
            # No predefined select: adjust the where clause.
            dummy, default_tables = self._find_spec.get_columns_and_tables()
            where = [Eq(*pair) for pair in zip(columns, values)]
            if self._where is not Undef:
                where.append(self._where)
            select = Select(1, And(*where), self._tables,
                            default_tables)
        else:
            # Rewrite the predefined query and use it as a subquery.
            aliased_columns = [Alias(column, "_key%d" % index)
                               for (index, column) in enumerate(columns)]
            subquery = replace_columns(self._get_select(), aliased_columns)
            where = [Eq(*pair) for pair in zip(aliased_columns, values)]
            select = Select(1, And(*where), Alias(subquery, "_tmp"))

        result = self._store._connection.execute(select)
        return result.get_one() is not None

    def is_empty(self):
        """Return C{True} if this result set doesn't contain any results."""
        subselect = self._get_select()
        subselect.limit = 1
        subselect.order_by = Undef
        select = Select(1, tables=Alias(subselect, "_tmp"), limit=1)
        result = self._store._connection.execute(select)
        return (not result.get_one())

    def any(self):
        """Return a single item from the result set.

        @return: An arbitrary object or C{None} if one isn't available.
        @see: L{one}, L{first}, and L{last}.
        """
        select = self._get_select()
        select.limit = 1
        select.order_by = Undef
        result = self._store._connection.execute(select)
        values = result.get_one()
        if values:
            return self._load_objects(result, values)
        return None

    def _any(self):
        """Return a single item from the result without changing sort order.

        @return: An arbitrary object or C{None} if one isn't available.
        """
        select = self._get_select()
        select.limit = 1
        result = self._store._connection.execute(select)
        values = result.get_one()
        if values:
            return self._load_objects(result, values)
        return None

    def first(self):
        """Return the first item from an ordered result set.

        @raises UnorderedError: Raised if the result set isn't ordered.
        @return: The first object or C{None} if one isn't available.
        @see: L{last}, L{one}, and L{any}.
        """
        if self._order_by is Undef:
            raise UnorderedError("Can't use first() on unordered result set")
        return self._any()

    def last(self):
        """Return the last item from an ordered result set.

        @raises FeatureError: Raised if the result set has a C{LIMIT} set.
        @raises UnorderedError: Raised if the result set isn't ordered.
        @return: The last object or C{None} if one isn't available.
        @see: L{first}, L{one}, and L{any}.
        """
        if self._order_by is Undef:
            raise UnorderedError("Can't use last() on unordered result set")
        if self._limit is not Undef:
            raise FeatureError("Can't use last() with a slice "
                               "of defined stop index")
        select = self._get_select()
        select.offset = Undef
        select.limit = 1
        select.order_by = []
        for expr in self._order_by:
            if isinstance(expr, Desc):
                select.order_by.append(expr.expr)
            elif isinstance(expr, Asc):
                select.order_by.append(Desc(expr.expr))
            else:
                select.order_by.append(Desc(expr))
        result = self._store._connection.execute(select)
        values = result.get_one()
        if values:
            return self._load_objects(result, values)
        return None

    def one(self):
        """Return one item from a result set containing at most one item.

        @raises NotOneError: Raised if the result set contains more than one
            item.
        @return: The object or C{None} if one isn't available.
        @see: L{first}, L{last}, and L{any}.
        """
        select = self._get_select()
        # limit could be 1 due to slicing, for instance.
        if select.limit is not Undef and select.limit > 2:
            select.limit = 2
        result = self._store._connection.execute(select)
        values = result.get_one()
        if result.get_one():
            raise NotOneError("one() used with more than one result available")
        if values:
            return self._load_objects(result, values)
        return None

    def order_by(self, *args):
        """Specify the ordering of the results.

        The query will be modified appropriately with an ORDER BY clause.

        Ascending and descending order can be specified by wrapping
        the columns in L{Asc} and L{Desc}.

        @param args: One or more L{storm.expr.Column} objects.
        """
        if self._offset is not Undef or self._limit is not Undef:
            raise FeatureError("Can't reorder a sliced result set")
        self._order_by = args or Undef
        return self

    def remove(self):
        """Remove all rows represented by this ResultSet from the database.

        This is done efficiently with a DELETE statement, so objects
        are not actually loaded into Python.
        """
        if self._group_by is not Undef:
            raise FeatureError("Removing isn't supported after a "
                               " GROUP BY clause ")
        if self._offset is not Undef or self._limit is not Undef:
            raise FeatureError("Can't remove a sliced result set")
        if self._find_spec.default_cls_info is None:
            raise FeatureError("Removing not yet supported for tuple or "
                               "expression finds")
        if self._select is not Undef:
            raise FeatureError("Removing isn't supported with "
                               "set expressions (unions, etc)")
        result = self._store._connection.execute(
            Delete(self._where, self._find_spec.default_cls_info.table))
        return result.rowcount

    def group_by(self, *expr):
        """Group this ResultSet by the given expressions.

        @param expr: The expressions used in the GROUP BY statement.

        @return: self (not a copy).
        """
        if self._select is not Undef:
            raise FeatureError("Grouping isn't supported with "
                               "set expressions (unions, etc)")

        find_spec = FindSpec(expr)
        columns, dummy = find_spec.get_columns_and_tables()

        self._group_by = columns
        return self

    def having(self, *expr):
        """Filter result previously grouped by.

        @param expr: Instances of L{Expr}.

        @return: self (not a copy).
        """
        if self._group_by is Undef:
            raise FeatureError("having can only be called after group_by.")
        self._having = And(*expr)
        return self

    def _aggregate(self, aggregate_func, expr, column=None):
        if self._group_by is not Undef:
            raise FeatureError("Single aggregates aren't supported after a "
                               " GROUP BY clause ")
        columns, default_tables = self._find_spec.get_columns_and_tables()
        if (self._select is Undef and not self._distinct and
            self._offset is Undef and self._limit is Undef):
            select = Select(aggregate_func(expr), self._where,
                            self._tables, default_tables)
        else:
            if expr is Undef:
                aggregate = aggregate_func(expr)
            else:
                alias = Alias(expr, "_expr")
                columns.append(alias)
                aggregate = aggregate_func(alias)
            # Ordering probably doesn't matter for any aggregates, and since
            # replace_columns() blows up on an ordered query, we'll drop it.
            select = self._get_select()
            select.order_by = Undef
            subquery = replace_columns(select, columns)
            select = Select(aggregate, tables=Alias(subquery, "_tmp"))
        result = self._store._connection.execute(select)
        value = result.get_one()[0]
        variable_factory = getattr(column, "variable_factory", None)
        if variable_factory:
            variable = variable_factory(allow_none=True)
            result.set_variable(variable, value)
            return variable.get()
        return value

    def count(self, expr=Undef, distinct=False):
        """Get the number of objects represented by this ResultSet."""
        return int(self._aggregate(lambda expr: Count(expr, distinct), expr))

    def max(self, expr):
        """Get the highest value from an expression."""
        return self._aggregate(Max, expr, expr)

    def min(self, expr):
        """Get the lowest value from an expression."""
        return self._aggregate(Min, expr, expr)

    def avg(self, expr):
        """Get the average value from an expression."""
        value = self._aggregate(Avg, expr)
        if value is None:
            return value
        return float(value)

    def sum(self, expr):
        """Get the sum of all values in an expression."""
        return self._aggregate(Sum, expr, expr)

    def get_select_expr(self, *columns):
        """Get a L{Select} expression to retrieve only the specified columns.

        @param columns: One or more L{storm.expr.Column} objects whose values
            will be fetched.
        @raises FeatureError: Raised if no columns are specified or if this
            result is a set expression such as a union.
        @return: A L{Select} expression configured to use the query parameters
            specified for this result set, and also limited to only retrieving
            data for the specified columns.
        """
        if not columns:
            raise FeatureError("select() takes at least one column "
                               "as argument")
        if self._select is not Undef:
            raise FeatureError(
                "Can't generate subselect expression for set expressions")
        select = self._get_select()
        select.columns = columns
        return select

    def values(self, *columns):
        """Retrieve only the specified columns.

        This does not load full objects from the database into Python.

        @param columns: One or more L{storm.expr.Column} objects whose
            values will be fetched.
        @raises FeatureError: Raised if no columns are specified or if this
            result is a set expression such as a union.
        @return: An iterator of tuples of the values for each column
            from each matching row in the database.
        """
        if not columns:
            raise FeatureError("values() takes at least one column "
                               "as argument")
        if self._select is not Undef:
            raise FeatureError("values() can't be used with set expressions")
        select = self._get_select()
        select.columns = columns
        result = self._store._connection.execute(select)
        if len(columns) == 1:
            variable = columns[0].variable_factory()
            for values in result:
                result.set_variable(variable, values[0])
                yield variable.get()
        else:
            variables = [column.variable_factory() for column in columns]
            for values in result:
                for variable, value in zip(variables, values):
                    result.set_variable(variable, value)
                yield tuple(variable.get() for variable in variables)

    def set(self, *args, **kwargs):
        """Update objects in the result set with the given arguments.

        This method will update all objects in the current result set
        to match expressions given as equalities or keyword arguments.
        These objects may still be in the database (an UPDATE is issued)
        or may be cached.

        For instance, C{result.set(Class.attr1 == 1, attr2=2)} will set
        C{attr1} to 1 and C{attr2} to 2, on all matching objects.
        """
        if self._group_by is not Undef:
            raise FeatureError("Setting isn't supported after a "
                               " GROUP BY clause ")

        if self._find_spec.default_cls_info is None:
            raise FeatureError("Setting isn't supported with tuple or "
                               "expression finds")
        if self._select is not Undef:
            raise FeatureError("Setting isn't supported with "
                               "set expressions (unions, etc)")

        if not (args or kwargs):
            return

        changes = {}
        cls = self._find_spec.default_cls_info.cls

        # For now only "Class.attr == var" is supported in args.
        for expr in args:
            if not isinstance(expr, Eq):
                raise FeatureError("Unsupported set expression: %r" %
                                   repr(expr))
            elif not isinstance(expr.expr1, Column):
                raise FeatureError("Unsupported left operand in set "
                                   "expression: %r" % repr(expr.expr1))
            elif not isinstance(expr.expr2, (Expr, Variable)):
                raise FeatureError("Unsupported right operand in set "
                                   "expression: %r" % repr(expr.expr2))
            changes[expr.expr1] = expr.expr2

        for key, value in kwargs.items():
            column = getattr(cls, key)
            if value is None:
                changes[column] = None
            elif isinstance(value, Expr):
                changes[column] = value
            else:
                changes[column] = column.variable_factory(value=value)

        expr = Update(changes, self._where,
                      self._find_spec.default_cls_info.table)
        self._store.execute(expr, noresult=True)

        try:
            cached = self.cached()
        except CompileError:
            # We are iterating through all objects in memory here, so
            # check if the object type matches to avoid trying to
            # invalidate a column that does not exist, on an unrelated
            # object.
            for obj_info in self._store._iter_alive():
                if obj_info.cls_info is self._find_spec.default_cls_info:
                    for column in changes:
                        obj_info.variables[column].set(AutoReload)
        else:
            changes = list(changes.items())
            for obj in cached:
                for column, value in changes:
                    variables = get_obj_info(obj).variables
                    if value is None:
                        pass
                    elif isinstance(value, Variable):
                        value = value.get()
                    elif isinstance(value, Expr):
                        # If the value is an Expression that means we
                        # can't compute it by ourselves: we rely on
                        # the database to compute it, so just set the
                        # value to AutoReload.
                        value = AutoReload
                    else:
                        value = variables[value].get()
                    variables[column].set(value)
                    variables[column].checkpoint()

    def cached(self):
        """Return matching objects from the cache for the current query."""
        if self._find_spec.default_cls_info is None:
            raise FeatureError("Cache finds not supported with tuples "
                               "or expressions")
        if self._tables is not Undef:
            raise FeatureError("Cache finds not supported with custom tables")
        if self._where is Undef:
            match = None
        else:
            match = compile_python.get_matcher(self._where)

            def get_column(column):
                return obj_info.variables[column].get()

        objects = []
        for obj_info in self._store._iter_alive():
            try:
                if (obj_info.cls_info is self._find_spec.default_cls_info and
                    (match is None or match(get_column))):
                    objects.append(self._store._get_object(obj_info))
            except LostObjectError:
                pass # This may happen when resolving lazy values
                     # in get_column().
        return objects

    def find(self, *args, **kwargs):
        """Perform a query on objects within this result set.

        This is analogous to L{Store.find}, although it doesn't take a
        C{cls_spec} argument, instead using the same tables as the
        existing result set, and restricts the results to those in
        this set.

        @param args: Instances of L{Expr}.
        @param kwargs: Mapping of simple column names to values or
            expressions to query for.

        @return: A L{ResultSet} of matching instances.
        """
        if self._select is not Undef:
            raise FeatureError("Can't query set expressions")
        if self._offset is not Undef or self._limit is not Undef:
            raise FeatureError("Can't query a sliced result set")
        if self._group_by is not Undef:
            raise FeatureError("Can't query grouped result sets")

        result_set = self.copy()
        extra_where = get_where_for_args(
            args, kwargs, self._find_spec.default_cls)
        if extra_where is not Undef:
            if result_set._where is Undef:
                result_set._where = extra_where
            else:
                result_set._where = And(result_set._where, extra_where)
        return result_set

    def _set_expr(self, expr_cls, other, all=False):
        if not self._find_spec.is_compatible(other._find_spec):
            raise FeatureError("Incompatible results for set operation")

        expr = expr_cls(self._get_select(), other._get_select(), all=all)
        return ResultSet(self._store, self._find_spec, select=expr)

    def union(self, other, all=False):
        """Get the L{Union} of this result set and another.

        @param all: If True, include duplicates.
        """
        if isinstance(other, EmptyResultSet):
            return self
        return self._set_expr(Union, other, all)

    def difference(self, other, all=False):
        """Get the difference, using L{Except}, of this result set and another.

        @param all: If True, include duplicates.
        """
        if isinstance(other, EmptyResultSet):
            return self
        return self._set_expr(Except, other, all)

    def intersection(self, other, all=False):
        """Get the L{Intersection} of this result set and another.

        @param all: If True, include duplicates.
        """
        if isinstance(other, EmptyResultSet):
            return other
        return self._set_expr(Intersect, other, all)


class EmptyResultSet:
    """An object that looks like a L{ResultSet} but represents no rows.

    This is convenient for application developers who want to provide
    a method which is guaranteed to return a L{ResultSet}-like object
    but which, in certain cases, knows there is no point in querying
    the database. For example::

        def get_people(self, ids):
            if not ids:
                return EmptyResultSet()
            return store.find(People, People.id.is_in(ids))

    The methods on EmptyResultSet (L{one}, L{config}, L{union}, etc)
    are meant to emulate a L{ResultSet} which has matched no rows.
    """

    def __init__(self, ordered=False):
        self._order_by = ordered

    def copy(self):
        result = EmptyResultSet(self._order_by)
        return result

    def config(self, distinct=None, offset=None, limit=None):
        return self

    def __iter__(self):
        return
        yield None

    def __len__(self):
        return 0

    def __getitem__(self, index):
        return self.copy()

    def __contains__(self, item):
        return False

    def is_empty(self):
        return True

    def any(self):
        return None

    def first(self):
        if self._order_by:
            return None
        raise UnorderedError("Can't use first() on unordered result set")

    def last(self):
        if self._order_by:
            return None
        raise UnorderedError("Can't use last() on unordered result set")

    def one(self):
        return None

    def order_by(self, *args):
        self._order_by = True
        return self

    def group_by(self, *expr):
        return self

    def remove(self):
        return 0

    def count(self, expr=Undef, distinct=False):
        return 0

    def max(self, column):
        return None

    def min(self, column):
        return None

    def avg(self, column):
        return None

    def sum(self, column):
        return None

    def get_select_expr(self, *columns):
        """Get a L{Select} expression to retrieve only the specified columns.

        @param columns: One or more L{storm.expr.Column} objects whose values
            will be fetched.
        @raises FeatureError: Raised if no columns are specified.
        @return: A L{Select} expression configured to use the query parameters
            specified for this result set.  The result of the select will
            always be an empty set of rows.
        """
        if not columns:
            raise FeatureError("select() takes at least one column "
                               "as argument")
        return Select(columns, False)

    def values(self, *columns):
        if not columns:
            raise FeatureError("values() takes at least one column "
                               "as argument")
        return
        yield None

    def set(self, *args, **kwargs):
        pass

    def cached(self):
        return []

    def find(self, *args, **kwargs):
        return self

    def union(self, other):
        if isinstance(other, EmptyResultSet):
            return self
        return other.union(self)

    def difference(self, other):
        return self

    def intersection(self, other):
        return self


class TableSet:
    """The representation of a set of tables which can be queried at once.

    This will typically be constructed by a call to L{Store.using}.
    """

    def __init__(self, store, tables):
        self._store = store
        self._tables = tables

    def find(self, cls_spec, *args, **kwargs):
        """Perform a query on the previously specified tables.

        This is identical to L{Store.find} except that the tables are
        explicitly specified instead of relying on inference.

        @return: A L{ResultSet}.
        """
        if self._store._implicit_flush_block_count == 0:
            self._store.flush()
        find_spec = FindSpec(cls_spec)
        where = get_where_for_args(args, kwargs, find_spec.default_cls)
        return self._store._result_set_factory(self._store, find_spec,
                                               where, self._tables)


Store._result_set_factory = ResultSet
Store._table_set = TableSet


class FindSpec:
    """The set of tables or expressions in the result of L{Store.find}."""

    def __init__(self, cls_spec):
        self.is_tuple = type(cls_spec) == tuple
        if not self.is_tuple:
            cls_spec = (cls_spec,)

        info = []
        for item in cls_spec:
            if isinstance(item, Expr):
                info.append((True, item))
            else:
                info.append((False, get_cls_info(item)))
        self._cls_spec_info = tuple(info)

        # Do we have a single non-expression item here?
        if not self.is_tuple and not info[0][0]:
            self.default_cls = cls_spec[0]
            self.default_cls_info = info[0][1]
            self.default_order = self.default_cls_info.default_order
        else:
            self.default_cls = None
            self.default_cls_info = None
            self.default_order = Undef

    def get_columns_and_tables(self):
        columns = []
        default_tables = []
        for is_expr, info in self._cls_spec_info:
            if is_expr:
                columns.append(info)
                if isinstance(info, Column):
                    default_tables.append(info.table)
            else:
                columns.extend(info.columns)
                default_tables.append(info.table)
        return columns, default_tables

    def is_compatible(self, find_spec):
        """Return True if this FindSpec is compatible with a second one.

        Two FindSpecs are considered compatible if either the find specs
        are identical (i.e. specifies the same classes and columns) or
        the find spec columns are of the same type.
        """
        if self.is_tuple != find_spec.is_tuple:
            return False
        if len(self._cls_spec_info) != len(find_spec._cls_spec_info):
            return False
        for (is_expr1, info1), (is_expr2, info2) in zip(
            self._cls_spec_info, find_spec._cls_spec_info):
            if is_expr1 != is_expr2:
                return False
            # If both infos are PropertyColumns, check whether they are
            # of the same type. Ideally we should check that the types as
            # defined in the database are the same, but checking the
            # variable class is easier and will work most of the time.
            if isinstance(info1, PropertyColumn):
                if not isinstance(info2, PropertyColumn):
                    return False
                variable_class1 = info1.variable_factory().__class__
                variable_class2 = info2.variable_factory().__class__
                if variable_class1 is not variable_class2:
                    return False
            elif info1 is not info2:
                return False
        return True

    def load_objects(self, store, result, values):
        objects = []
        values_start = values_end = 0
        for is_expr, info in self._cls_spec_info:
            if is_expr:
                values_end += 1
                variable = getattr(info, "variable_factory", Variable)(
                    value=values[values_start], from_db=True)
                objects.append(variable.get())
            else:
                values_end += len(info.columns)
                obj = store._load_object(info, result,
                                         values[values_start:values_end])
                objects.append(obj)
            values_start = values_end
        if self.is_tuple:
            return tuple(objects)
        else:
            return objects[0]

    def get_columns_and_values_for_item(self, item):
        """Generate a comparison expression with the given item."""
        if isinstance(item, tuple):
            if not self.is_tuple:
                raise TypeError("Find spec does not expect tuples.")
        else:
            if self.is_tuple:
                raise TypeError("Find spec expects tuples.")
            item = (item,)

        columns = []
        values = []
        for (is_expr, info), value in zip(self._cls_spec_info, item):
            if is_expr:
                if not isinstance(value, (Expr, Variable)) and (
                    value is not None):
                    value = getattr(info, "variable_factory", Variable)(
                        value=value)
                columns.append(info)
                values.append(value)
            else:
                obj_info = get_obj_info(value)
                if obj_info.cls_info != info:
                    raise TypeError("%r does not match %r" % (value, info))
                columns.extend(info.primary_key)
                values.extend(obj_info.primary_vars)
        return columns, values


def get_where_for_args(args, kwargs, cls=None):
    equals = list(args)
    if kwargs:
        if cls is None:
            raise FeatureError("Can't determine class that keyword "
                               "arguments are associated with")
        for key, value in kwargs.items():
            equals.append(getattr(cls, key) == value)
    if equals:
        return And(*equals)
    return Undef


def replace_columns(expr, columns):
    if isinstance(expr, Select):
        select = copy(expr)
        select.columns = columns
        # Remove the ordering if it won't affect the result of the query.
        if select.limit is Undef and select.offset is Undef:
            select.order_by = Undef
        return select
    elif isinstance(expr, SetExpr):
        # The ORDER BY clause might refer to columns we have replaced.
        # Luckily we can ignore it if there is no limit/offset.
        if expr.order_by is not Undef and (
            expr.limit is not Undef or expr.offset is not Undef):
            raise FeatureError(
                "__contains__() does not yet support set "
                "expressions that combine ORDER BY with "
                "LIMIT/OFFSET")
        subexprs = [replace_columns(subexpr, columns)
                    for subexpr in expr.exprs]
        return expr.__class__(
            all=expr.all, limit=expr.limit, offset=expr.offset,
            *subexprs)
    else:
        raise FeatureError(
            "__contains__() does not yet support %r expressions"
            % (expr.__class__,))


class AutoReload(LazyValue):
    """A marker for reloading a single value.

    Often this will be used to specify that a specific attribute
    should be loaded from the database on the next access, like so::

        storm_object.property = AutoReload

    On the next access to C{storm_object.property}, the value will be
    loaded from the database.

    It is also often used as a default value for a property::

        class Person(object):
            __storm_table__ = "person"
            id = Int(allow_none=False, default=AutoReload)

        person = store.add(Person)
        person.id # gets the attribute from the database.
    """
    pass

AutoReload = AutoReload()


class block_access:
    """
    Context manager blocks database access by one or more L{Store}\\ s in the
    managed scope.
    """

    def __init__(self, *args):
        self.stores = args

    def __enter__(self):
        for store in self.stores:
            store.block_access()

    def __exit__(self, exc_type, exc_val, exc_tb):
        for store in self.stores:
            store.unblock_access()



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from datetime import timedelta
import time
import os

from storm.exceptions import OperationalError
from storm.databases.sqlite import SQLite
from storm.database import create_database
from storm.uri import URI
from storm.tests.databases.base import DatabaseTest, UnsupportedDatabaseTest
from storm.tests.helper import TestHelper, MakePath


class SQLiteMemoryTest(DatabaseTest, TestHelper):

    helpers = [MakePath]
    
    def get_path(self):
        return ""

    def create_database(self):
        self.database = SQLite(URI("sqlite:%s?synchronous=OFF&timeout=0" %
                                   self.get_path()))

    def create_tables(self):
        self.connection.execute("CREATE TABLE number "
                                "(one INTEGER, two INTEGER, three INTEGER)")
        self.connection.execute("CREATE TABLE test "
                                "(id INTEGER PRIMARY KEY, title VARCHAR)")
        self.connection.execute("CREATE TABLE datetime_test "
                                "(id INTEGER PRIMARY KEY,"
                                " dt TIMESTAMP, d DATE, t TIME, td INTERVAL)")
        self.connection.execute("CREATE TABLE bin_test "
                                "(id INTEGER PRIMARY KEY, b BLOB)")

    def drop_tables(self):
        pass

    def test_wb_create_database(self):
        database = create_database("sqlite:")
        self.assertTrue(isinstance(database, SQLite))
        self.assertEqual(database._filename, ":memory:")

    def test_concurrent_behavior(self):
        pass # We can't connect to the in-memory database twice, so we can't
             # exercise the concurrency behavior (nor it makes sense).

    def test_synchronous(self):
        synchronous_values = {"OFF": 0, "NORMAL": 1, "FULL": 2}
        for value in synchronous_values:
            database = SQLite(URI("sqlite:%s?synchronous=%s" %
                                  (self.get_path(), value)))
            connection = database.connect()
            result = connection.execute("PRAGMA synchronous")
            self.assertEqual(result.get_one()[0],
                             synchronous_values[value])

    def test_sqlite_specific_reserved_words(self):
        """Check sqlite-specific reserved words are recognized.

        This uses a list copied from http://www.sqlite.org/lang_keywords.html
        with the reserved words from SQL1992 removed.
        """
        reserved_words = """
            abort after analyze attach autoincrement before conflict
            database detach each exclusive explain fail glob if ignore
            index indexed instead isnull limit notnull offset plan
            pragma query raise regexp reindex release rename replace
            row savepoint temp trigger vacuum virtual
            """.split()
        for word in reserved_words:
            self.assertTrue(self.connection.compile.is_reserved_word(word),
                            "Word missing: %s" % (word,))


class SQLiteFileTest(SQLiteMemoryTest):

    def get_path(self):
        return self.make_path()

    def test_wb_create_database(self):
        filename = self.make_path()
        database = create_database("sqlite:%s" % filename)
        self.assertTrue(isinstance(database, SQLite))
        self.assertEqual(database._filename, filename)

    def test_timeout(self):
        database = create_database("sqlite:%s?timeout=0.3" % self.get_path())
        connection1 = database.connect()
        connection2 = database.connect()
        connection1.execute("CREATE TABLE test (id INTEGER PRIMARY KEY)")
        connection1.commit()
        connection1.execute("INSERT INTO test VALUES (1)")
        started = time.time()
        try:
            connection2.execute("INSERT INTO test VALUES (2)")
        except OperationalError as exception:
            self.assertEqual(str(exception), "database is locked")
            self.assertTrue(time.time()-started >= 0.3)
        else:
            self.fail("OperationalError not raised")

    def test_commit_timeout(self):
        """Regression test for commit observing the timeout.
        
        In 0.10, the timeout wasn't observed for connection.commit().

        """
        # Create a database with a table.
        database = create_database("sqlite:%s?timeout=0.3" % self.get_path())
        connection1 = database.connect()
        connection1.execute("CREATE TABLE test (id INTEGER PRIMARY KEY)")
        connection1.commit()

        # Put some data in, but also make a second connection to the database,
        # which will prevent a commit until it is closed.
        connection1.execute("INSERT INTO test VALUES (1)")
        connection2 = database.connect()
        connection2.execute("SELECT id FROM test")

        started = time.time()
        try:
            connection1.commit()
        except OperationalError as exception:
            self.assertEqual(str(exception), "database is locked")
            # In 0.10, the next assertion failed because the timeout wasn't
            # enforced for the "COMMIT" statement.
            self.assertTrue(time.time()-started >= 0.3)
        else:
            self.fail("OperationalError not raised")

    def test_recover_after_timeout(self):
        """Regression test for recovering from database locked exception.
        
        In 0.10, connection.commit() would forget that a transaction was in
        progress if an exception was raised, such as an OperationalError due to
        another connection being open.  As a result, a subsequent modification
        to the database would cause BEGIN to be issued to the database, which
        would complain that a transaction was already in progress.

        """
        # Create a database with a table.
        database = create_database("sqlite:%s?timeout=0.3" % self.get_path())
        connection1 = database.connect()
        connection1.execute("CREATE TABLE test (id INTEGER PRIMARY KEY)")
        connection1.commit()

        # Put some data in, but also make a second connection to the database,
        # which will prevent a commit until it is closed.
        connection1.execute("INSERT INTO test VALUES (1)")
        connection2 = database.connect()
        connection2.execute("SELECT id FROM test")
        self.assertRaises(OperationalError, connection1.commit)

        # Close the second connection - it should now be possible to commit.
        connection2.close()

        # In 0.10, the next statement raised OperationalError: cannot start a
        # transaction within a transaction
        connection1.execute("INSERT INTO test VALUES (2)")
        connection1.commit()

        # Check that the correct data is present
        self.assertEqual(connection1.execute("SELECT id FROM test").get_all(),
                         [(1,), (2,)])

    def test_journal(self):
        journal_values = {"DELETE": 'delete', "TRUNCATE": 'truncate',
                          "PERSIST": 'persist', "MEMORY": 'memory',
                          "WAL": 'wal', "OFF": 'off'}
        for value in journal_values:
            database = SQLite(URI("sqlite:%s?journal_mode=%s" %
                                  (self.get_path(), value)))
            connection = database.connect()
            result = connection.execute("PRAGMA journal_mode").get_one()[0]
            self.assertEqual(result,
                             journal_values[value])

    def test_journal_persistency_to_rollback(self):
        journal_values = {"DELETE": 'delete', "TRUNCATE": 'truncate',
                          "PERSIST": 'persist', "MEMORY": 'memory',
                          "WAL": 'wal', "OFF": 'off'}
        for value in journal_values:
            database = SQLite(URI("sqlite:%s?journal_mode=%s" %
                                  (self.get_path(), value)))
            connection = database.connect()
            connection.execute("CREATE TABLE test (id INTEGER PRIMARY KEY)")
            connection.rollback()
            result = connection.execute("PRAGMA journal_mode").get_one()[0]
            self.assertEqual(result,
                             journal_values[value])

    def test_foreign_keys(self):
        foreign_keys_values = {"ON": 1, "OFF": 0}
        for value in foreign_keys_values:
            database = SQLite(URI("sqlite:%s?foreign_keys=%s" %
                                  (self.get_path(), value)))
            connection = database.connect()
            result = connection.execute("PRAGMA foreign_keys").get_one()[0]
            self.assertEqual(result,
                             foreign_keys_values[value])

    def test_foreign_keys_persistency_to_rollback(self):
        foreign_keys_values = {"ON": 1, "OFF": 0}
        for value in foreign_keys_values:
            database = SQLite(URI("sqlite:%s?foreign_keys=%s" %
                                  (self.get_path(), value)))
            connection = database.connect()
            connection.execute("CREATE TABLE test (id INTEGER PRIMARY KEY)")
            connection.rollback()
            result = connection.execute("PRAGMA foreign_keys").get_one()[0]
            self.assertEqual(result,
                             foreign_keys_values[value])

class SQLiteUnsupportedTest(UnsupportedDatabaseTest, TestHelper):
 
    dbapi_module_names = ["pysqlite2", "sqlite3"]
    db_module_name = "sqlite"



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from urllib.parse import quote

from storm.exceptions import URIError


class URI:
    """A representation of a Uniform Resource Identifier (URI).

    This is intended exclusively for database connection URIs.

    @ivar username: The username part of the URI, or C{None}.
    @ivar password: The password part of the URI, or C{None}.
    @ivar host: The host part of the URI, or C{None}.
    @type port: L{int}
    @ivar port: The port part of the URI, or C{None}.
    @ivar database: The part of the URI representing the database name, or
        C{None}.
    """

    username = None
    password = None
    host = None
    port = None
    database = None

    def __init__(self, uri_str):
        try:
            self.scheme, rest = uri_str.split(":", 1)
        except ValueError:
            raise URIError("URI has no scheme: %s" % repr(uri_str))

        self.options = {}

        if "?" in rest:
            rest, options = rest.split("?", 1)
            for pair in options.split("&"):
                key, value = pair.split("=", 1)
                self.options[unescape(key)] = unescape(value)
        if rest:
            if not rest.startswith("//"):
                self.database = unescape(rest)
            else:
                rest = rest[2:]
                if "/" in rest:
                    rest, database = rest.split("/", 1)
                    self.database = unescape(database)
                if "@" in rest:
                    userpass, hostport = rest.split("@", 1)
                else:
                    userpass = None
                    hostport = rest
                if hostport:
                    if ":" in hostport:
                        host, port = hostport.rsplit(":", 1)
                        self.host = unescape(host)
                        if port:
                            self.port = int(port)
                    else:
                        self.host = unescape(hostport)
                if userpass is not None:
                    if ":" in userpass:
                        username, password = userpass.rsplit(":", 1)
                        self.username = unescape(username)
                        self.password = unescape(password)
                    else:
                        self.username = unescape(userpass)

    def copy(self):
        uri = object.__new__(self.__class__)
        uri.__dict__.update(self.__dict__)
        uri.options = self.options.copy()
        return uri

    def __str__(self):
        tokens = [self.scheme, ":"]
        append = tokens.append
        if (self.username is not None or self.password is not None or
            self.host is not None or self.port is not None):
            append("//")
            if self.username is not None or self.password is not None:
                if self.username is not None:
                    append(escape(self.username))
                if self.password is not None:
                    append(":")
                    append(escape(self.password))
                append("@")
            if self.host is not None:
                append(escape(self.host))
            if self.port is not None:
                append(":")
                append(str(self.port))
            append("/")
        if self.database is not None:
            append(escape(self.database, "/"))
        if self.options:
            options = ["%s=%s" % (escape(key), escape(value))
                       for key, value in sorted(self.options.items())]
            append("?")
            append("&".join(options))
        return "".join(tokens)


def escape(s, safe=""):
    return quote(s, safe)


def unescape(s):
    if "%" not in s:
        return s
    i = 0
    j = s.find("%")
    r = []
    while j != -1:
        r.append(s[i:j])
        i = j+3
        r.append(chr(int(s[j+1:i], 16)))
        j = s.find("%", i)
    r.append(s[i:])
    return "".join(r)



import datetime
import os
import sys
from unittest import TestCase

from storm.tests import has_fixtures

# Optional dependency. If missing, Fixture tests are skipped.
if has_fixtures:
    import fixtures.testcase
    TestWithFixtures = fixtures.testcase.TestWithFixtures
    from storm.testing import CaptureTracer
else:
    TestWithFixtures = object

try:
    # Optional dependency, if missing TimelineTracer tests are skipped.
    import timeline
    has_timeline = True
except ImportError:
    has_timeline = False

from storm.tracer import (trace, install_tracer, get_tracers, remove_tracer,
                          remove_tracer_type, remove_all_tracers, debug,
                          BaseStatementTracer, DebugTracer, TimeoutTracer,
                          TimelineTracer, TimeoutError, _tracers)
from storm.database import Connection, create_database
from storm.expr import Variable
from storm.tests.helper import TestHelper


class TracerTest(TestHelper):

    def tearDown(self):
        super().tearDown()
        del _tracers[:]

    def test_install_tracer(self):
        c = object()
        d = object()
        install_tracer(c)
        install_tracer(d)
        self.assertEqual(get_tracers(), [c, d])

    def test_remove_all_tracers(self):
        install_tracer(object())
        remove_all_tracers()
        self.assertEqual(get_tracers(), [])

    def test_remove_tracer(self):
        """The C{remote_tracer} function removes a specific tracer."""
        tracer1 = object()
        tracer2 = object()
        install_tracer(tracer1)
        install_tracer(tracer2)
        remove_tracer(tracer1)
        self.assertEqual(get_tracers(), [tracer2])

    def test_remove_tracer_with_not_installed_tracer(self):
        """C{remote_tracer} exits gracefully if the tracer is not installed."""
        tracer = object()
        remove_tracer(tracer)
        self.assertEqual(get_tracers(), [])

    def test_remove_tracer_type(self):
        class C:
            pass

        class D(C):
            pass

        c = C()
        d1 = D()
        d2 = D()
        install_tracer(d1)
        install_tracer(c)
        install_tracer(d2)
        remove_tracer_type(C)
        self.assertEqual(get_tracers(), [d1, d2])
        remove_tracer_type(D)
        self.assertEqual(get_tracers(), [])

    def test_install_debug(self):
        debug(True)
        debug(True)
        self.assertEqual([type(x) for x in get_tracers()], [DebugTracer])

    def test_wb_install_debug_with_custom_stream(self):
        marker = object()
        debug(True, marker)
        [tracer] = get_tracers()
        self.assertEqual(tracer._stream, marker)

    def test_remove_debug(self):
        debug(True)
        debug(True)
        debug(False)
        self.assertEqual(get_tracers(), [])

    def test_trace(self):
        stash = []

        class Tracer:
            def m1(_, *args, **kwargs):
                stash.extend(["m1", args, kwargs])

            def m2(_, *args, **kwargs):
                stash.extend(["m2", args, kwargs])

        install_tracer(Tracer())
        trace("m1", 1, 2, c=3)
        trace("m2")
        trace("m3")
        self.assertEqual(stash, ["m1", (1, 2), {"c": 3}, "m2", (), {}])


class MockVariable(Variable):

    def __init__(self, value):
        self._value = value

    def get(self, to_db=False):
        return self._value


class DebugTracerTest(TestHelper):

    def setUp(self):
        super().setUp()
        self.stream = self.mocker.mock(type(sys.stderr))
        self.tracer = DebugTracer(self.stream)

        datetime_mock = self.mocker.replace("datetime.datetime")
        datetime_mock.now()
        self.mocker.result(datetime.datetime(1, 2, 3, 4, 5, 6, 7))
        self.mocker.count(0, 1)

        self.variable = MockVariable("PARAM")

    def tearDown(self):
        del _tracers[:]
        super().tearDown()

    def test_wb_debug_tracer_uses_stderr_by_default(self):
        self.mocker.replay()

        tracer = DebugTracer()
        self.assertEqual(tracer._stream, sys.stderr)

    def test_wb_debug_tracer_uses_first_arg_as_stream(self):
        self.mocker.replay()

        marker = object()
        tracer = DebugTracer(marker)
        self.assertEqual(tracer._stream, marker)

    def test_connection_raw_execute(self):
        self.stream.write(
            "[04:05:06.000007] EXECUTE: 'STATEMENT', ('PARAM',)\n")
        self.stream.flush()
        self.mocker.replay()

        connection = "CONNECTION"
        raw_cursor = "RAW_CURSOR"
        statement = "STATEMENT"
        params = [self.variable]

        self.tracer.connection_raw_execute(connection, raw_cursor,
                                           statement, params)

    def test_connection_raw_execute_with_non_variable(self):
        self.stream.write(
            "[04:05:06.000007] EXECUTE: 'STATEMENT', ('PARAM', 1)\n")
        self.stream.flush()
        self.mocker.replay()

        connection = "CONNECTION"
        raw_cursor = "RAW_CURSOR"
        statement = "STATEMENT"
        params = [self.variable, 1]

        self.tracer.connection_raw_execute(connection, raw_cursor,
                                           statement, params)

    def test_connection_raw_execute_error(self):
        self.stream.write("[04:05:06.000007] ERROR: ERROR\n")
        self.stream.flush()
        self.mocker.replay()

        connection = "CONNECTION"
        raw_cursor = "RAW_CURSOR"
        statement = "STATEMENT"
        params = "PARAMS"
        error = "ERROR"

        self.tracer.connection_raw_execute_error(connection, raw_cursor,
                                                 statement, params, error)

    def test_connection_raw_execute_success(self):
        self.stream.write("[04:05:06.000007] DONE\n")
        self.stream.flush()
        self.mocker.replay()

        connection = "CONNECTION"
        raw_cursor = "RAW_CURSOR"
        statement = "STATEMENT"
        params = "PARAMS"

        self.tracer.connection_raw_execute_success(connection, raw_cursor,
                                                   statement, params)

    def test_connection_commit(self):
        self.stream.write("[04:05:06.000007] COMMIT xid=None\n")
        self.stream.flush()
        self.mocker.replay()

        connection = "CONNECTION"

        self.tracer.connection_commit(connection)

    def test_connection_rollback(self):
        self.stream.write("[04:05:06.000007] ROLLBACK xid=None\n")
        self.stream.flush()
        self.mocker.replay()

        connection = "CONNECTION"

        self.tracer.connection_rollback(connection)


class TimeoutTracerTestBase(TestHelper):

    tracer_class = TimeoutTracer

    def setUp(self):
        super().setUp()
        self.tracer = self.tracer_class()
        self.raw_cursor = self.mocker.mock()
        self.statement = self.mocker.mock()
        self.params = self.mocker.mock()

        # Some data is kept in the connection, so we use a proxy to
        # allow things we don't care about here to happen.
        class Connection:
            pass

        self.connection = self.mocker.proxy(Connection())

    def tearDown(self):
        super().tearDown()
        del _tracers[:]

    def execute(self):
        self.tracer.connection_raw_execute(self.connection, self.raw_cursor,
                                           self.statement, self.params)

    def execute_raising(self):
        self.assertRaises(TimeoutError, self.tracer.connection_raw_execute,
                          self.connection, self.raw_cursor,
                          self.statement, self.params)


class TimeoutTracerTest(TimeoutTracerTestBase):

    def test_raise_not_implemented(self):
        """
        L{TimeoutTracer.connection_raw_execute_error},
        L{TimeoutTracer.set_statement_timeout} and
        L{TimeoutTracer.get_remaining_time} must all be implemented by
        backend-specific subclasses.
        """
        self.assertRaises(NotImplementedError,
                          self.tracer.connection_raw_execute_error,
                          None, None, None, None, None)
        self.assertRaises(NotImplementedError,
                          self.tracer.set_statement_timeout, None, None)
        self.assertRaises(NotImplementedError,
                          self.tracer.get_remaining_time)

    def test_raise_timeout_error_when_no_remaining_time(self):
        """
        A L{TimeoutError} is raised if there isn't any time left when a
        statement is executed.
        """
        tracer_mock = self.mocker.patch(self.tracer)
        tracer_mock.get_remaining_time()
        self.mocker.result(0)
        self.mocker.replay()

        try:
            self.execute()
        except TimeoutError as e:
            self.assertEqual("0 seconds remaining in time budget", e.message)
            self.assertEqual(self.statement, e.statement)
            self.assertEqual(self.params, e.params)
        else:
            self.fail("TimeoutError not raised")

    def test_raise_timeout_on_granularity(self):
        tracer_mock = self.mocker.patch(self.tracer)

        self.mocker.order()

        tracer_mock.get_remaining_time()
        self.mocker.result(self.tracer.granularity)
        tracer_mock.set_statement_timeout(self.raw_cursor,
                                          self.tracer.granularity)
        tracer_mock.get_remaining_time()
        self.mocker.result(0)
        self.mocker.replay()

        self.execute()
        self.execute_raising()

    def test_wont_raise_timeout_before_granularity(self):
        tracer_mock = self.mocker.patch(self.tracer)

        self.mocker.order()

        tracer_mock.get_remaining_time()
        self.mocker.result(self.tracer.granularity)
        tracer_mock.set_statement_timeout(self.raw_cursor,
                                          self.tracer.granularity)
        tracer_mock.get_remaining_time()
        self.mocker.result(1)
        self.mocker.replay()

        self.execute()
        self.execute()

    def test_always_set_when_remaining_time_increased(self):
        tracer_mock = self.mocker.patch(self.tracer)

        self.mocker.order()

        tracer_mock.get_remaining_time()
        self.mocker.result(1)
        tracer_mock.set_statement_timeout(self.raw_cursor, 1)
        tracer_mock.get_remaining_time()
        self.mocker.result(2)
        tracer_mock.set_statement_timeout(self.raw_cursor, 2)
        self.mocker.replay()

        self.execute()
        self.execute()

    def test_set_again_on_granularity(self):
        tracer_mock = self.mocker.patch(self.tracer)

        self.mocker.order()

        tracer_mock.get_remaining_time()
        self.mocker.result(self.tracer.granularity * 2)
        tracer_mock.set_statement_timeout(self.raw_cursor,
                                          self.tracer.granularity * 2)
        tracer_mock.get_remaining_time()
        self.mocker.result(self.tracer.granularity)
        tracer_mock.set_statement_timeout(self.raw_cursor,
                                          self.tracer.granularity)
        self.mocker.replay()

        self.execute()
        self.execute()

    def test_set_again_after_granularity(self):
        tracer_mock = self.mocker.patch(self.tracer)

        self.mocker.order()

        tracer_mock.get_remaining_time()
        self.mocker.result(self.tracer.granularity * 2)
        tracer_mock.set_statement_timeout(self.raw_cursor,
                                          self.tracer.granularity * 2)
        tracer_mock.get_remaining_time()
        self.mocker.result(self.tracer.granularity - 1)
        tracer_mock.set_statement_timeout(self.raw_cursor,
                                          self.tracer.granularity - 1)
        self.mocker.replay()

        self.execute()
        self.execute()


class TimeoutTracerWithDBTest(TestHelper):

    def setUp(self):
        super().setUp()
        self.tracer = StuckInTimeTimeoutTracer(10)
        install_tracer(self.tracer)
        database = create_database(os.environ["STORM_POSTGRES_URI"])
        self.connection = database.connect()

    def tearDown(self):
        super().tearDown()
        remove_tracer(self.tracer)
        self.connection.close()

    def is_supported(self):
        return bool(os.environ.get("STORM_POSTGRES_URI"))

    def test_timeout_set_on_beginning_of_new_transaction__commit(self):
        """Check that we set the statement timeout before the first query of a
        transaction regardless of the remaining time left by previous
        transactions.

        When we reuse a connection for a different transaction, the remaining
        time of a previous transaction (which is stored in the connection)
        could cause the first query in that transaction to run with no
        timeout. This test makes sure that doesn't happen.
        """
        self.connection.execute('SELECT 1')
        self.assertEqual([10], self.tracer.set_statement_timeout_calls)

        self.connection.commit()

        self.connection.execute('SELECT 1')
        self.assertEqual([10, 10], self.tracer.set_statement_timeout_calls)

    def test_timeout_set_on_beginning_of_new_transaction__rollback(self):
        """Same as the test above, but here we rollback the first tx."""
        self.connection.execute('SELECT 1')
        self.assertEqual([10], self.tracer.set_statement_timeout_calls)

        self.connection.rollback()

        self.connection.execute('SELECT 1')
        self.assertEqual([10, 10], self.tracer.set_statement_timeout_calls)


class StuckInTimeTimeoutTracer(TimeoutTracer):

    def __init__(self, fixed_remaining_time):
        super().__init__()
        self.set_statement_timeout_calls = []
        self.fixed_remaining_time = fixed_remaining_time

    def get_remaining_time(self):
        return self.fixed_remaining_time

    def set_statement_timeout(self, raw_cursor, remaining_time):
        self.set_statement_timeout_calls.append(remaining_time)


class StubConnection(Connection):

    def __init__(self):
        self._database = None
        self._event = None
        self._raw_connection = None
        self.name = 'Foo'


class BaseStatementTracerTest(TestCase):

    class LoggingBaseStatementTracer(BaseStatementTracer):
        def _expanded_raw_execute(self, connection, raw_cursor, statement):
            self.__dict__.setdefault('calls', []).append(
                (connection, raw_cursor, statement))

    def test_no_params(self):
        """With no parameters the statement is passed through verbatim."""
        tracer = self.LoggingBaseStatementTracer()
        tracer.connection_raw_execute('foo', 'bar', 'baz ? %s', ())
        self.assertEqual([('foo', 'bar', 'baz ? %s')], tracer.calls)

    def test_params_substituted_pyformat(self):
        tracer = self.LoggingBaseStatementTracer()
        conn = StubConnection()
        conn.param_mark = '%s'
        var1 = MockVariable('VAR1')
        tracer.connection_raw_execute(
            conn, 'cursor', 'SELECT * FROM person where name = %s', [var1])
        self.assertEqual(
            [(conn, 'cursor', "SELECT * FROM person where name = 'VAR1'")],
            tracer.calls)

    def test_params_substituted_single_string(self):
        """String parameters are formatted as a single quoted string."""
        tracer = self.LoggingBaseStatementTracer()
        conn = StubConnection()
        var1 = MockVariable('VAR1')
        tracer.connection_raw_execute(
            conn, 'cursor', 'SELECT * FROM person where name = ?', [var1])
        self.assertEqual(
            [(conn, 'cursor', "SELECT * FROM person where name = 'VAR1'")],
            tracer.calls)

    def test_qmark_percent_s_literal_preserved(self):
        """With ? parameters %s in the statement can be kept intact."""
        tracer = self.LoggingBaseStatementTracer()
        conn = StubConnection()
        var1 = MockVariable(1)
        tracer.connection_raw_execute(
            conn, 'cursor',
            "SELECT * FROM person where id > ? AND name LIKE '%s'", [var1])
        self.assertEqual(
            [(conn, 'cursor',
              "SELECT * FROM person where id > 1 AND name LIKE '%s'")],
            tracer.calls)

    def test_int_variable_as_int(self):
        """Int parameters are formatted as an int literal."""
        tracer = self.LoggingBaseStatementTracer()
        conn = StubConnection()
        var1 = MockVariable(1)
        tracer.connection_raw_execute(
            conn, 'cursor', "SELECT * FROM person where id = ?", [var1])
        self.assertEqual(
            [(conn, 'cursor', "SELECT * FROM person where id = 1")],
            tracer.calls)

    def test_like_clause_preserved(self):
        """% operators in LIKE statements are preserved."""
        tracer = self.LoggingBaseStatementTracer()
        conn = StubConnection()
        var1 = MockVariable('substring')
        tracer.connection_raw_execute(
            conn, 'cursor',
            "SELECT * FROM person WHERE name LIKE '%%' || ? || '-suffix%%'",
            [var1])
        self.assertEqual(
            [(conn, 'cursor', "SELECT * FROM person WHERE name "
                              "LIKE '%%' || 'substring' || '-suffix%%'")],
            tracer.calls)

    def test_unformattable_statements_are_handled(self):
        tracer = self.LoggingBaseStatementTracer()
        conn = StubConnection()
        var1 = MockVariable('substring')
        tracer.connection_raw_execute(
            conn, 'cursor', "%s %s",
            [var1])
        self.assertEqual(
            [(conn, 'cursor',
              "Unformattable query: '%%s %%s' with params [%r]." %
              'substring')],
            tracer.calls)


class TimelineTracerTest(TestHelper):

    def is_supported(self):
        return has_timeline

    def factory(self):
        self.timeline = timeline.Timeline()
        return self.timeline

    def test_separate_tracers_own_state(self):
        """Check that multiple TimelineTracer's could be used at once."""
        tracer1 = TimelineTracer(self.factory)
        tracer2 = TimelineTracer(self.factory)
        tracer1.threadinfo.action = 'foo'
        self.assertEqual(None, getattr(tracer2.threadinfo, 'action', None))

    def test_error_finishes_action(self):
        tracer = TimelineTracer(self.factory)
        action = timeline.Timeline().start('foo', 'bar')
        tracer.threadinfo.action = action
        tracer.connection_raw_execute_error(
            'conn', 'cursor', 'statement', 'params', 'error')
        self.assertNotEqual(None, action.duration)

    def test_success_finishes_action(self):
        tracer = TimelineTracer(self.factory)
        action = timeline.Timeline().start('foo', 'bar')
        tracer.threadinfo.action = action
        tracer.connection_raw_execute_success(
            'conn', 'cursor', 'statement', 'params')
        self.assertNotEqual(None, action.duration)

    def test_finds_timeline_from_factory(self):
        factory_result = timeline.Timeline()
        tracer = TimelineTracer(lambda: factory_result)
        tracer._expanded_raw_execute('conn', 'cursor', 'statement')
        self.assertEqual(1, len(factory_result.actions))

    def test_action_details_are_statement(self):
        """The detail in the timeline action is the formatted SQL statement."""
        tracer = TimelineTracer(self.factory)
        tracer._expanded_raw_execute('conn', 'cursor', 'statement')
        self.assertEqual('statement', self.timeline.actions[-1].detail)

    def test_category_from_prefix_and_connection_name(self):
        tracer = TimelineTracer(self.factory, prefix='bar-')
        tracer._expanded_raw_execute(StubConnection(), 'cursor', 'statement')
        self.assertEqual('bar-Foo', self.timeline.actions[-1].category)

    def test_unnamed_connection(self):
        """A connection with no name has <unknown> put in as a placeholder."""
        tracer = TimelineTracer(self.factory, prefix='bar-')
        tracer._expanded_raw_execute('conn', 'cursor', 'statement')
        self.assertEqual('bar-<unknown>', self.timeline.actions[-1].category)

    def test_default_prefix(self):
        """By default the prefix "SQL-" is added to the action's category."""
        tracer = TimelineTracer(self.factory)
        tracer._expanded_raw_execute('conn', 'cursor', 'statement')
        self.assertEqual('SQL-<unknown>', self.timeline.actions[-1].category)


class CaptureTracerTest(TestHelper, TestWithFixtures):

    def is_supported(self):
        return has_fixtures

    def tearDown(self):
        super().tearDown()
        del _tracers[:]

    def test_capture(self):
        """
        Using the L{CaptureTracer} fixture starts capturing queries and stops
        removes the tracer upon cleanup.
        """
        tracer = self.useFixture(CaptureTracer())
        self.assertEqual([tracer], get_tracers())
        conn = StubConnection()
        conn.param_mark = '%s'
        var = MockVariable("var")
        tracer.connection_raw_execute(conn, "cursor", "select %s", [var])
        self.assertEqual(["select 'var'"], tracer.queries)

        def check():
            self.assertEqual([], get_tracers())

        self.addCleanup(check)

    def test_capture_as_context_manager(self):
        """{CaptureTracer}s can be used as context managers."""
        conn = StubConnection()
        with CaptureTracer() as tracer:
            self.assertEqual([tracer], get_tracers())
            tracer.connection_raw_execute(conn, "cursor", "select", [])
        self.assertEqual([], get_tracers())
        self.assertEqual(["select"], tracer.queries)

    def test_capture_multiple(self):
        """L{CaptureTracer}s can be used as nested context managers."""

        conn = StubConnection()

        def trace(statement):
            for tracer in get_tracers():
                tracer.connection_raw_execute(conn, "cursor", statement, [])

        with CaptureTracer() as tracer1:
            trace("one")
            with CaptureTracer() as tracer2:
                trace("two")
            trace("three")

        self.assertEqual([], get_tracers())
        self.assertEqual(["one", "two", "three"], tracer1.queries)
        self.assertEqual(["two"], tracer2.queries)

    def test_capture_with_exception(self):
        """
        L{CaptureTracer}s re-raise any error when used as context managers.
        """
        errors = []
        try:
            with CaptureTracer():
                raise RuntimeError("boom")
        except RuntimeError as error:
            errors.append(error)
        [error] = errors
        self.assertEqual("boom", str(error))
        self.assertEqual([], get_tracers())



import time
import random
import transaction as zope_transaction

from functools import wraps

from zope.component import getUtility

from storm.zope.interfaces import IZStorm
from storm.exceptions import IntegrityError, DisconnectionError

from twisted.internet.threads import deferToThreadPool


RETRIABLE_ERRORS = (DisconnectionError, IntegrityError)
try:
    from psycopg2.extensions import TransactionRollbackError
    RETRIABLE_ERRORS = RETRIABLE_ERRORS + (TransactionRollbackError,)
except ImportError:
    pass


class Transactor:
    """Run in a thread code that needs to interact with the database.

    This class makes sure that code interacting with the database is run
    in a separate thread and that the associated transaction is aborted or
    committed in the same thread.

    @param threadpool: The C{ThreadPool} to get threads from.
    @param _transaction: The C{TransactionManager} to use, for test cases only.

    @ivar retries: Maximum number of retries upon retriable exceptions. The
        default is to retry a function up to 2 times upon possibly transient
        or spurious errors like L{IntegrityError} and L{DisconnectionError}.
    @ivar on_retry: If not C{None}, a callable that will be called before
        retrying to run a function, and passed a L{RetryContext} instance with
        the details about the retry.

    @see: C{twisted.python.threadpool.ThreadPool}
    """
    retries = 2
    on_retry = None

    sleep = time.sleep
    uniform = random.uniform

    def __init__(self, threadpool, transaction=None):
        self._threadpool = threadpool
        if transaction is None:
            transaction = zope_transaction
        self._transaction = transaction

    def run(self, function, *args, **kwargs):
        """Run C{function} in a thread.

        The function is run in a thread by a function wrapper, which
        commits the transaction if the function runs successfully. If it
        raises an exception the transaction is aborted.

        @param function: The function to run.
        @param args: Positional arguments to pass to C{function}.
        @param kwargs: Keyword arguments to pass to C{function}.
        @return: A C{Deferred} that will fire after the function has been run.
        """
        # Inline the reactor import here for sake of safeness, in case a
        # custom reactor needs to be installed
        from twisted.internet import reactor
        return deferToThreadPool(
            reactor, self._threadpool, self._wrap, function, *args, **kwargs)

    def _wrap(self, function, *args, **kwargs):
        retries = 0
        while True:
            try:
                result = function(*args, **kwargs)
                self._transaction.commit()
            except RETRIABLE_ERRORS as error:
                if isinstance(error, DisconnectionError):
                    # If we got a disconnection, calling rollback may not be
                    # enough because psycopg2 doesn't necessarily use the
                    # connection, so we call a dummy query to be sure that all
                    # the stores are correct.
                    zstorm = getUtility(IZStorm)
                    for name, store in zstorm.iterstores():
                        try:
                            store.execute("SELECT 1")
                        except DisconnectionError:
                            pass
                self._transaction.abort()
                if retries < self.retries:
                    retries += 1
                    if self.on_retry is not None:
                        context = RetryContext(function, args, kwargs, retries,
                                               error)
                        self.on_retry(context)
                    self.sleep(self.uniform(1, 2 ** retries))
                    continue
                else:
                    raise
            except:
                self._transaction.abort()
                raise
            else:
                return result


class RetryContext:
    """Hold details about a function that is going to be retried.

    @ivar function: The function that is going to be retried.
    @ivar args: The positional arguments passed to the function.
    @ivar kwargs: The keyword arguments passed to the function.
    @ivar retry: The sequential number of the retry that is going to be
        performed.
    @ivar error: The Exception instance that caused a retry to be scheduled.
    """

    def __init__(self, function, args, kwargs, retry, error):
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.retry = retry
        self.error = error


def transact(method):
    """Decorate L{method} so that it is invoked via L{Transactor.run}.

    Example::

        from twisted.python.threadpool import ThreadPool
        from storm.twisted.transact import Transactor, transact

        class Foo(object):

            def __init__(self, transactor):
                self.transactor = transactor

            @transact
            def bar(self):
                # code that uses Storm

        threadpool = ThreadPool(0, 10)
        threadpool.start()
        transactor = Transactor(threadpool)
        foo = Foo(transactor)
        deferred = foo.bar()
        deferred.addCallback(...)

    @param method: The method to decorate.
    @return: A decorated method.

    @note: The return value of the decorated method should *not* contain
        any reference to Storm objects, because they were retrieved in
        a different thread and cannot be used outside it.
    """

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        return self.transactor.run(method, self, *args, **kwargs)
    return wrapper



#
# Copyright (c) 2011 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = [
    'find_tests',
    'has_fixtures',
    'has_psycopg',
    'has_subunit',
    ]

import doctest
from itertools import chain
import os
import unittest

try:
    import fixtures
    fixtures  # Silence lint.
except ImportError:
    has_fixtures = False
else:
    has_fixtures = True


try:
    import psycopg2
    psycopg2  # Silence lint.
except ImportError:
    has_psycopg = False
else:
    has_psycopg = True


try:
    import subunit
    subunit  # Silence lint.
except ImportError:
    has_subunit = False
else:
    has_subunit = True


def find_tests(testpaths=()):
    """Find all test paths, or test paths contained in the provided sequence.

    @param testpaths: If provided, only tests in the given sequence will
                      be considered.  If not provided, all tests are
                      considered.
    @return: a test suite containing the requested tests.
    """
    suite = unittest.TestSuite()
    topdir = os.path.abspath(
        os.path.join(os.path.dirname(__file__), os.pardir, os.pardir))
    testdir = os.path.dirname(__file__)
    docdir = os.path.join(os.path.dirname(testdir), "docs")
    testpaths = set(testpaths)
    for root, dirnames, filenames in chain(os.walk(testdir), os.walk(docdir)):
        for filename in filenames:
            filepath = os.path.join(root, filename)
            relpath = filepath[len(topdir)+1:]

            if (filename == "__init__.py" or filename.endswith(".pyc") or
                    relpath == os.path.join("storm", "docs", "conf.py") or
                    relpath == os.path.join("storm", "tests", "conftest.py")):
                # Skip non-tests.
                continue

            if testpaths:
                # Skip any tests not in testpaths.
                for testpath in testpaths:
                    if relpath.startswith(testpath):
                        break
                else:
                    continue

            if filename.endswith(".py"):
                modpath = relpath.replace(os.path.sep, ".")[:-3]
                module = __import__(modpath, None, None, [""])
                suite.addTest(
                    unittest.defaultTestLoader.loadTestsFromModule(module))
            elif filename.endswith(".rst"):
                load_test = True
                if relpath == os.path.join("storm", "docs", "zope.rst"):
                    # Special case the inclusion of the Zope-dependent
                    # ZStorm doctest.
                    import storm.tests.zope as ztest
                    load_test = (
                        ztest.has_transaction and
                        ztest.has_zope_component and
                        ztest.has_zope_security)
                if load_test:
                    parent_path = os.path.dirname(relpath).replace(
                        os.path.sep, ".")
                    parent_module = __import__(parent_path, None, None, [""])
                    suite.addTest(doctest.DocFileSuite(
                            os.path.basename(relpath),
                            module_relative=True,
                            package=parent_module,
                            optionflags=doctest.ELLIPSIS))

    return suite



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from datetime import datetime, date, time, timedelta
from decimal import Decimal as decimal
import gc
import json
import uuid

from storm.exceptions import NoneError, PropertyPathError
from storm.properties import PropertyPublisherMeta
from storm.properties import *
from storm.variables import *
from storm.info import get_obj_info
from storm.expr import Column, Select, compile, State, SQLRaw
from storm.tests.info import Wrapper
from storm.tests.helper import TestHelper


class CustomVariable(Variable):
    pass

class Custom(SimpleProperty):
    variable_class = CustomVariable


class PropertyTest(TestHelper):

    def setUp(self):
        TestHelper.setUp(self)
        class Class:
            __storm_table__ = "mytable"
            prop1 = Custom("column1", primary=True)
            prop2 = Custom()
            prop3 = Custom("column3", default=50, allow_none=False)
        class SubClass(Class):
            __storm_table__ = "mysubtable"
        self.Class = Class
        self.SubClass = SubClass

    def test_column(self):
        self.assertTrue(isinstance(self.Class.prop1, Column))

    def test_cls(self):
        self.assertEqual(self.Class.prop1.cls, self.Class)
        self.assertEqual(self.Class.prop2.cls, self.Class)
        self.assertEqual(self.SubClass.prop1.cls, self.SubClass)
        self.assertEqual(self.SubClass.prop2.cls, self.SubClass)
        self.assertEqual(self.Class.prop1.cls, self.Class)
        self.assertEqual(self.Class.prop2.cls, self.Class)

    def test_cls_reverse(self):
        self.assertEqual(self.SubClass.prop1.cls, self.SubClass)
        self.assertEqual(self.SubClass.prop2.cls, self.SubClass)
        self.assertEqual(self.Class.prop1.cls, self.Class)
        self.assertEqual(self.Class.prop2.cls, self.Class)
        self.assertEqual(self.SubClass.prop1.cls, self.SubClass)
        self.assertEqual(self.SubClass.prop2.cls, self.SubClass)

    def test_name(self):
        self.assertEqual(self.Class.prop1.name, "column1")

    def test_auto_name(self):
        self.assertEqual(self.Class.prop2.name, "prop2")

    def test_auto_table(self):
        self.assertEqual(self.Class.prop1.table, self.Class)
        self.assertEqual(self.Class.prop2.table, self.Class)

    def test_auto_table_subclass(self):
        self.assertEqual(self.Class.prop1.table, self.Class)
        self.assertEqual(self.Class.prop2.table, self.Class)
        self.assertEqual(self.SubClass.prop1.table, self.SubClass)
        self.assertEqual(self.SubClass.prop2.table, self.SubClass)

    def test_auto_table_subclass_reverse_initialization(self):
        self.assertEqual(self.SubClass.prop1.table, self.SubClass)
        self.assertEqual(self.SubClass.prop2.table, self.SubClass)
        self.assertEqual(self.Class.prop1.table, self.Class)
        self.assertEqual(self.Class.prop2.table, self.Class)

    def test_variable_factory(self):
        variable = self.Class.prop1.variable_factory()
        self.assertTrue(isinstance(variable, CustomVariable))
        self.assertFalse(variable.is_defined())

        variable = self.Class.prop3.variable_factory()
        self.assertTrue(isinstance(variable, CustomVariable))
        self.assertTrue(variable.is_defined())

    def test_variable_factory_validator_attribute(self):
        # Should work even if we make things harder by reusing properties.
        prop = Custom()
        class Class1:
            __storm_table__ = "table1"
            prop1 = prop
        class Class2:
            __storm_table__ = "table2"
            prop2 = prop
        args = []
        def validator(obj, attr, value):
            args.append((obj, attr, value))
        variable1 = Class1.prop1.variable_factory(validator=validator)
        variable2 = Class2.prop2.variable_factory(validator=validator)
        variable1.set(1)
        variable2.set(2)
        self.assertEqual(args, [(None, "prop1", 1), (None, "prop2", 2)])

    def test_default(self):
        obj = self.SubClass()
        self.assertEqual(obj.prop1, None)
        self.assertEqual(obj.prop2, None)
        self.assertEqual(obj.prop3, 50)

    def test_set_get(self):
        obj = self.Class()
        obj.prop1 = 10
        obj.prop2 = 20
        obj.prop3 = 30
        self.assertEqual(obj.prop1, 10)
        self.assertEqual(obj.prop2, 20)
        self.assertEqual(obj.prop3, 30)

    def test_set_get_none(self):
        obj = self.Class()
        obj.prop1 = None
        obj.prop2 = None
        self.assertEqual(obj.prop1, None)
        self.assertEqual(obj.prop2, None)
        self.assertRaises(NoneError, setattr, obj, "prop3", None)

    def test_set_with_validator(self):
        args = []
        def validator(obj, attr, value):
            args[:] = obj, attr, value
            return 42

        class Class:
            __storm_table__ = "mytable"
            prop = Custom("column", primary=True, validator=validator)

        obj = Class()
        obj.prop = 21

        self.assertEqual(args, [obj, "prop", 21])
        self.assertEqual(obj.prop, 42)

    def test_set_get_subclass(self):
        obj = self.SubClass()
        obj.prop1 = 10
        obj.prop2 = 20
        obj.prop3 = 30
        self.assertEqual(obj.prop1, 10)
        self.assertEqual(obj.prop2, 20)
        self.assertEqual(obj.prop3, 30)

    def test_set_get_explicitly(self):
        obj = self.Class()
        prop1 = self.Class.prop1
        prop2 = self.Class.prop2
        prop3 = self.Class.prop3
        prop1.__set__(obj, 10)
        prop2.__set__(obj, 20)
        prop3.__set__(obj, 30)
        self.assertEqual(prop1.__get__(obj), 10)
        self.assertEqual(prop2.__get__(obj), 20)
        self.assertEqual(prop3.__get__(obj), 30)

    def test_set_get_subclass_explicitly(self):
        obj = self.SubClass()
        prop1 = self.Class.prop1
        prop2 = self.Class.prop2
        prop3 = self.Class.prop3
        prop1.__set__(obj, 10)
        prop2.__set__(obj, 20)
        prop3.__set__(obj, 30)
        self.assertEqual(prop1.__get__(obj), 10)
        self.assertEqual(prop2.__get__(obj), 20)
        self.assertEqual(prop3.__get__(obj), 30)

    def test_delete(self):
        obj = self.Class()
        obj.prop1 = 10
        obj.prop2 = 20
        obj.prop3 = 30
        del obj.prop1
        del obj.prop2
        del obj.prop3
        self.assertEqual(obj.prop1, None)
        self.assertEqual(obj.prop2, None)
        self.assertEqual(obj.prop3, None)

    def test_delete_subclass(self):
        obj = self.SubClass()
        obj.prop1 = 10
        obj.prop2 = 20
        obj.prop3 = 30
        del obj.prop1
        del obj.prop2
        del obj.prop3
        self.assertEqual(obj.prop1, None)
        self.assertEqual(obj.prop2, None)
        self.assertEqual(obj.prop3, None)

    def test_delete_explicitly(self):
        obj = self.Class()
        obj.prop1 = 10
        obj.prop2 = 20
        obj.prop3 = 30
        self.Class.prop1.__delete__(obj)
        self.Class.prop2.__delete__(obj)
        self.Class.prop3.__delete__(obj)
        self.assertEqual(obj.prop1, None)
        self.assertEqual(obj.prop2, None)
        self.assertEqual(obj.prop3, None)

    def test_delete_subclass_explicitly(self):
        obj = self.SubClass()
        obj.prop1 = 10
        obj.prop2 = 20
        obj.prop3 = 30
        self.Class.prop1.__delete__(obj)
        self.Class.prop2.__delete__(obj)
        self.Class.prop3.__delete__(obj)
        self.assertEqual(obj.prop1, None)
        self.assertEqual(obj.prop2, None)
        self.assertEqual(obj.prop3, None)

    def test_comparable_expr(self):
        prop1 = self.Class.prop1
        prop2 = self.Class.prop2
        prop3 = self.Class.prop3
        expr = Select(SQLRaw("*"), (prop1 == "value1") &
                                   (prop2 == "value2") &
                                   (prop3 == "value3"))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "SELECT * FROM mytable WHERE "
                                    "mytable.column1 = ? AND "
                                    "mytable.prop2 = ? AND "
                                    "mytable.column3 = ?")

        self.assertVariablesEqual(
            state.parameters,
            [CustomVariable("value1"),
             CustomVariable("value2"),
             CustomVariable("value3")])

    def test_comparable_expr_subclass(self):
        prop1 = self.SubClass.prop1
        prop2 = self.SubClass.prop2
        prop3 = self.SubClass.prop3
        expr = Select(SQLRaw("*"), (prop1 == "value1") &
                                   (prop2 == "value2") &
                                   (prop3 == "value3"))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "SELECT * FROM mysubtable WHERE "
                                    "mysubtable.column1 = ? AND "
                                    "mysubtable.prop2 = ? AND "
                                    "mysubtable.column3 = ?")
        self.assertVariablesEqual(
            state.parameters,
            [CustomVariable("value1"),
             CustomVariable("value2"),
             CustomVariable("value3")])

    def test_set_get_delete_with_wrapper(self):
        obj = self.Class()
        get_obj_info(obj) # Ensure the obj_info exists for obj.
        self.Class.prop1.__set__(Wrapper(obj), 10)
        self.assertEqual(self.Class.prop1.__get__(Wrapper(obj)), 10)
        self.Class.prop1.__delete__(Wrapper(obj))
        self.assertEqual(self.Class.prop1.__get__(Wrapper(obj)), None)

    def test_reuse_of_instance(self):
        """Properties are dynamically bound to the class where they're used.

        It basically means that the property may be instantiated
        independently from the class itself, and reused in any number of
        classes.  It's not something we should announce as granted, but
        right now it works, and we should try not to break it.
        """
        prop = Custom()
        class Class1:
            __storm_table__ = "table1"
            prop1 = prop
        class Class2:
            __storm_table__ = "table2"
            prop2 = prop
        self.assertEqual(Class1.prop1.name, "prop1")
        self.assertEqual(Class1.prop1.table, Class1)
        self.assertEqual(Class2.prop2.name, "prop2")
        self.assertEqual(Class2.prop2.table, Class2)


class PropertyKindsTest(TestHelper):

    def setup(self, property, *args, **kwargs):
        prop2_kwargs = kwargs.pop("prop2_kwargs", {})
        kwargs["primary"] = True
        class Class:
            __storm_table__ = "mytable"
            prop1 = property("column1", *args, **kwargs)
            prop2 = property(**prop2_kwargs)
        class SubClass(Class):
            pass
        self.Class = Class
        self.SubClass = SubClass
        self.obj = SubClass()
        self.obj_info = get_obj_info(self.obj)
        self.column1 = self.SubClass.prop1
        self.column2 = self.SubClass.prop2
        self.variable1 = self.obj_info.variables[self.column1]
        self.variable2 = self.obj_info.variables[self.column2]

    def test_bool(self):
        self.setup(Bool, default=True, allow_none=False)

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, BoolVariable))
        self.assertTrue(isinstance(self.variable2, BoolVariable))

        self.assertEqual(self.obj.prop1, True)
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.obj.prop1 = 1
        self.assertTrue(self.obj.prop1 is True)
        self.obj.prop1 = 0
        self.assertTrue(self.obj.prop1 is False)

    def test_int(self):
        self.setup(Int, default=50, allow_none=False)

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, IntVariable))
        self.assertTrue(isinstance(self.variable2, IntVariable))

        self.assertEqual(self.obj.prop1, 50)
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.obj.prop1 = False
        self.assertEqual(self.obj.prop1, 0)
        self.obj.prop1 = True
        self.assertEqual(self.obj.prop1, 1)

    def test_float(self):
        self.setup(Float, default=50.5, allow_none=False)

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, FloatVariable))
        self.assertTrue(isinstance(self.variable2, FloatVariable))

        self.assertEqual(self.obj.prop1, 50.5)
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.obj.prop1 = 1
        self.assertTrue(isinstance(self.obj.prop1, float))

    def test_decimal(self):
        self.setup(Decimal, default=decimal("50.5"), allow_none=False)

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, DecimalVariable))
        self.assertTrue(isinstance(self.variable2, DecimalVariable))

        self.assertEqual(self.obj.prop1, decimal("50.5"))
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.obj.prop1 = 1
        self.assertTrue(isinstance(self.obj.prop1, decimal))

    def test_bytes(self):
        self.setup(Bytes, default=b"def", allow_none=False)

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, BytesVariable))
        self.assertTrue(isinstance(self.variable2, BytesVariable))

        self.assertEqual(self.obj.prop1, b"def")
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.assertRaises(TypeError, setattr, self.obj, "prop1", "unicode")

    def test_unicode(self):
        self.setup(Unicode, default="def", allow_none=False)

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, UnicodeVariable))
        self.assertTrue(isinstance(self.variable2, UnicodeVariable))

        self.assertEqual(self.obj.prop1, "def")
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.assertRaises(TypeError, setattr, self.obj, "prop1", b"str")

    def test_datetime(self):
        self.setup(DateTime, default=0, allow_none=False)

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, DateTimeVariable))
        self.assertTrue(isinstance(self.variable2, DateTimeVariable))

        self.assertEqual(self.obj.prop1, datetime.utcfromtimestamp(0))
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.obj.prop1 = 0.0
        self.assertEqual(self.obj.prop1, datetime.utcfromtimestamp(0))
        self.obj.prop1 = datetime(2006, 1, 1, 12, 34)
        self.assertEqual(self.obj.prop1, datetime(2006, 1, 1, 12, 34))

        self.assertRaises(TypeError, setattr, self.obj, "prop1", object())

    def test_date(self):
        self.setup(Date, default=date(2006, 1, 1), allow_none=False)

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, DateVariable))
        self.assertTrue(isinstance(self.variable2, DateVariable))

        self.assertEqual(self.obj.prop1, date(2006, 1, 1))
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.obj.prop1 = datetime(2006, 1, 1, 12, 34, 56)
        self.assertEqual(self.obj.prop1, date(2006, 1, 1))
        self.obj.prop1 = date(2006, 1, 1)
        self.assertEqual(self.obj.prop1, date(2006, 1, 1))

        self.assertRaises(TypeError, setattr, self.obj, "prop1", object())

    def test_time(self):
        self.setup(Time, default=time(12, 34), allow_none=False)

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, TimeVariable))
        self.assertTrue(isinstance(self.variable2, TimeVariable))

        self.assertEqual(self.obj.prop1, time(12, 34))
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.obj.prop1 = datetime(2006, 1, 1, 12, 34, 56)
        self.assertEqual(self.obj.prop1, time(12, 34, 56))
        self.obj.prop1 = time(12, 34, 56)
        self.assertEqual(self.obj.prop1, time(12, 34, 56))

        self.assertRaises(TypeError, setattr, self.obj, "prop1", object())

    def test_timedelta(self):
        self.setup(TimeDelta,
                   default=timedelta(days=1, seconds=2, microseconds=3),
                   allow_none=False)

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, TimeDeltaVariable))
        self.assertTrue(isinstance(self.variable2, TimeDeltaVariable))

        self.assertEqual(self.obj.prop1,
                         timedelta(days=1, seconds=2, microseconds=3))
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.obj.prop1 = timedelta(days=42, seconds=42, microseconds=42)
        self.assertEqual(self.obj.prop1,
                         timedelta(days=42, seconds=42, microseconds=42))

        self.assertRaises(TypeError, setattr, self.obj, "prop1", object())

    def test_uuid(self):
        value1 = uuid.UUID("{0609f76b-878f-4546-baf5-c1b135e8de72}")
        value2 = uuid.UUID("{c9703f9d-0abb-47d7-a793-8f90f1b98d5e}")
        self.setup(UUID, default=value1, allow_none=False)

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, UUIDVariable))
        self.assertTrue(isinstance(self.variable2, UUIDVariable))

        self.assertEqual(self.obj.prop1, value1)
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.obj.prop1 = value1
        self.assertEqual(self.obj.prop1, value1)
        self.obj.prop1 = value2
        self.assertEqual(self.obj.prop1, value2)

        self.assertRaises(TypeError, setattr, self.obj, "prop1",
                          "{0609f76b-878f-4546-baf5-c1b135e8de72}")

    def test_enum(self):
        self.setup(Enum, map={"foo": 1, "bar": 2},
                   default="foo", allow_none=False,
                   prop2_kwargs=dict(map={"foo": 1, "bar": 2}))

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, EnumVariable))
        self.assertTrue(isinstance(self.variable2, EnumVariable))

        self.assertEqual(self.obj.prop1, "foo")
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.obj.prop1 = "foo"
        self.assertEqual(self.obj.prop1, "foo")
        self.obj.prop1 = "bar"
        self.assertEqual(self.obj.prop1, "bar")

        self.assertRaises(ValueError, setattr, self.obj, "prop1", "baz")
        self.assertRaises(ValueError, setattr, self.obj, "prop1", 1)

    def test_enum_with_set_map(self):
        self.setup(Enum, map={"foo": 1, "bar": 2},
                   set_map={"fooics": 1, "barics": 2},
                   default="fooics", allow_none=False,
                   prop2_kwargs=dict(map={"foo": 1, "bar": 2}))

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, EnumVariable))
        self.assertTrue(isinstance(self.variable2, EnumVariable))

        self.assertEqual(self.obj.prop1, "foo")
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.obj.prop1 = "fooics"
        self.assertEqual(self.obj.prop1, "foo")
        self.obj.prop1 = "barics"
        self.assertEqual(self.obj.prop1, "bar")

        self.assertRaises(ValueError, setattr, self.obj, "prop1", "foo")
        self.assertRaises(ValueError, setattr, self.obj, "prop1", 1)

    def test_pickle(self):
        self.setup(Pickle, default_factory=dict, allow_none=False)

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, PickleVariable))
        self.assertTrue(isinstance(self.variable2, PickleVariable))

        self.assertEqual(self.obj.prop1, {})
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.obj.prop1 = []
        self.assertEqual(self.obj.prop1, [])
        self.obj.prop1.append("a")
        self.assertEqual(self.obj.prop1, ["a"])

    def test_pickle_events(self):
        self.setup(Pickle, default_factory=list, allow_none=False)

        changes = []
        def changed(owner, variable, old_value, new_value, fromdb):
            changes.append((variable, old_value, new_value, fromdb))

        # Can't checkpoint Undef.
        self.obj.prop2 = []

        self.obj_info.checkpoint()
        self.obj_info.event.emit("start-tracking-changes", self.obj_info.event)
        self.obj_info.event.hook("changed", changed)

        self.assertEqual(self.obj.prop1, [])
        self.assertEqual(changes, [])
        self.obj.prop1.append("a")
        self.assertEqual(changes, [])

        # Check "flush" event. Notice that the other variable wasn't
        # listed, since it wasn't changed.
        self.obj_info.event.emit("flush")
        self.assertEqual(changes, [(self.variable1, None, ["a"], False)])

        del changes[:]

        # Check "object-deleted" event. Notice that the other variable
        # wasn't listed again, since it wasn't changed.
        del self.obj
        self.assertEqual(changes, [(self.variable1, None, ["a"], False)])

    def test_json(self):
        # Skip test if json support is not available.
        if json is None:
            return

        self.setup(JSON, default_factory=dict, allow_none=False)

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, JSONVariable))
        self.assertTrue(isinstance(self.variable2, JSONVariable))

        self.assertEqual(self.obj.prop1, {})
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.obj.prop1 = []
        self.assertEqual(self.obj.prop1, [])
        self.obj.prop1.append("a")
        self.assertEqual(self.obj.prop1, ["a"])

    def test_json_events(self):
        # Skip test if json support is not available.
        if json is None:
            return

        self.setup(JSON, default_factory=list, allow_none=False)

        changes = []
        def changed(owner, variable, old_value, new_value, fromdb):
            changes.append((variable, old_value, new_value, fromdb))

        # Can't checkpoint Undef.
        self.obj.prop2 = []

        self.obj_info.checkpoint()
        self.obj_info.event.emit("start-tracking-changes", self.obj_info.event)
        self.obj_info.event.hook("changed", changed)

        self.assertEqual(self.obj.prop1, [])
        self.assertEqual(changes, [])
        self.obj.prop1.append("a")
        self.assertEqual(changes, [])

        # Check "flush" event. Notice that the other variable wasn't
        # listed, since it wasn't changed.
        self.obj_info.event.emit("flush")
        self.assertEqual(changes, [(self.variable1, None, ["a"], False)])

        del changes[:]

        # Check "object-deleted" event. Notice that the other variable
        # wasn't listed again, since it wasn't changed.
        del self.obj
        self.assertEqual(changes, [(self.variable1, None, ["a"], False)])

    def test_list(self):
        self.setup(List, default_factory=list, allow_none=False)

        self.assertTrue(isinstance(self.column1, Column))
        self.assertTrue(isinstance(self.column2, Column))
        self.assertEqual(self.column1.name, "column1")
        self.assertEqual(self.column1.table, self.SubClass)
        self.assertEqual(self.column2.name, "prop2")
        self.assertEqual(self.column2.table, self.SubClass)
        self.assertTrue(isinstance(self.variable1, ListVariable))
        self.assertTrue(isinstance(self.variable2, ListVariable))

        self.assertEqual(self.obj.prop1, [])
        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
        self.obj.prop2 = None
        self.assertEqual(self.obj.prop2, None)

        self.obj.prop1 = ["a"]
        self.assertEqual(self.obj.prop1, ["a"])
        self.obj.prop1.append("b")
        self.assertEqual(self.obj.prop1, ["a", "b"])

    def test_list_events(self):
        self.setup(List, default_factory=list, allow_none=False)

        changes = []
        def changed(owner, variable, old_value, new_value, fromdb):
            changes.append((variable, old_value, new_value, fromdb))

        self.obj_info.checkpoint()
        self.obj_info.event.emit("start-tracking-changes", self.obj_info.event)
        self.obj_info.event.hook("changed", changed)

        self.assertEqual(self.obj.prop1, [])
        self.assertEqual(changes, [])
        self.obj.prop1.append("a")
        self.assertEqual(changes, [])

        # Check "flush" event. Notice that the other variable wasn't
        # listed, since it wasn't changed.
        self.obj_info.event.emit("flush")
        self.assertEqual(changes, [(self.variable1, None, ["a"], False)])

        del changes[:]

        # Check "object-deleted" event. Notice that the other variable
        # wasn't listed again, since it wasn't changed.
        del self.obj
        self.assertEqual(changes, [(self.variable1, None, ["a"], False)])

    def test_variable_factory_arguments(self):
        class Class:
            __storm_table__ = "test"
            id = Int(primary=True)

        validator_args = []
        def validator(obj, attr, value):
            validator_args[:] = obj, attr, value
            return value

        for func, cls, value in [
                               (Bool, BoolVariable, True),
                               (Int, IntVariable, 1),
                               (Float, FloatVariable, 1.1),
                               (Bytes, BytesVariable, b"str"),
                               (Unicode, UnicodeVariable, "unicode"),
                               (DateTime, DateTimeVariable, datetime.now()),
                               (Date, DateVariable, date.today()),
                               (Time, TimeVariable, datetime.now().time()),
                               (Pickle, PickleVariable, {}),
                                     ]:

            # Test no default and allow_none=True.
            Class.prop = func(name="name")
            column = Class.prop.__get__(None, Class)
            self.assertEqual(column.name, "name")
            self.assertEqual(column.table, Class)

            variable = column.variable_factory()
            self.assertTrue(isinstance(variable, cls))
            self.assertEqual(variable.get(), None)
            variable.set(None)
            self.assertEqual(variable.get(), None)

            # Test default and allow_none=False.
            Class.prop = func(name="name", default=value, allow_none=False)
            column = Class.prop.__get__(None, Class)
            self.assertEqual(column.name, "name")
            self.assertEqual(column.table, Class)

            variable = column.variable_factory()
            self.assertTrue(isinstance(variable, cls))
            self.assertRaises(NoneError, variable.set, None)
            self.assertEqual(variable.get(), value)

            # Test default=None and allow_none=False (incoherent).
            Class.prop = func(name="name", default=None, allow_none=False)
            column = Class.prop.__get__(None, Class)
            self.assertEqual(column.name, "name")
            self.assertEqual(column.table, Class)

            self.assertRaises(NoneError, column.variable_factory)

            # Test default_factory.
            Class.prop = func(name="name", default_factory=lambda:value)
            column = Class.prop.__get__(None, Class)
            self.assertEqual(column.name, "name")
            self.assertEqual(column.table, Class)

            variable = column.variable_factory()
            self.assertTrue(isinstance(variable, cls))
            self.assertEqual(variable.get(), value)

            # Test validator.
            Class.prop = func(name="name", validator=validator, default=value)
            column = Class.prop.__get__(None, Class)
            self.assertEqual(column.name, "name")
            self.assertEqual(column.table, Class)

            del validator_args[:]
            variable = column.variable_factory()
            self.assertTrue(isinstance(variable, cls))
            # Validator is not called on instantiation.
            self.assertEqual(validator_args, [])
            # But is when setting the variable.
            variable.set(value)
            self.assertEqual(validator_args, [None, "prop", value])


class PropertyRegistryTest(TestHelper):

    def setUp(self):
        TestHelper.setUp(self)

        class Class:
            __storm_table__ = "mytable"
            prop1 = Property("column1", primary=True)
            prop2 = Property()

        class SubClass(Class):
            __storm_table__ = "mysubtable"

        self.Class = Class
        self.SubClass = SubClass
        self.AnotherClass = type("Class", (Class,), {})

        self.registry = PropertyRegistry()

    def test_get_empty(self):
        self.assertRaises(PropertyPathError, self.registry.get, "unexistent")

    def test_get(self):
        self.registry.add_class(self.Class)
        prop1 = self.registry.get("prop1")
        prop2 = self.registry.get("prop2")
        self.assertTrue(prop1 is self.Class.prop1)
        self.assertTrue(prop2 is self.Class.prop2)

    def test_get_with_class_name(self):
        self.registry.add_class(self.Class)
        prop1 = self.registry.get("Class.prop1")
        prop2 = self.registry.get("Class.prop2")
        self.assertTrue(prop1 is self.Class.prop1)
        self.assertTrue(prop2 is self.Class.prop2)

    def test_get_with_two_classes(self):
        self.registry.add_class(self.Class)
        self.registry.add_class(self.SubClass)
        prop1 = self.registry.get("Class.prop1")
        prop2 = self.registry.get("Class.prop2")
        self.assertTrue(prop1 is self.Class.prop1)
        self.assertTrue(prop2 is self.Class.prop2)
        prop1 = self.registry.get("SubClass.prop1")
        prop2 = self.registry.get("SubClass.prop2")
        self.assertTrue(prop1 is self.SubClass.prop1)
        self.assertTrue(prop2 is self.SubClass.prop2)

    def test_get_ambiguous(self):
        self.AnotherClass.__module__ += ".foo"
        self.registry.add_class(self.Class)
        self.registry.add_class(self.SubClass)
        self.registry.add_class(self.AnotherClass)
        self.assertRaises(PropertyPathError, self.registry.get, "Class.prop1")
        self.assertRaises(PropertyPathError, self.registry.get, "Class.prop2")
        prop1 = self.registry.get("SubClass.prop1")
        prop2 = self.registry.get("SubClass.prop2")
        self.assertTrue(prop1 is self.SubClass.prop1)
        self.assertTrue(prop2 is self.SubClass.prop2)

    def test_get_ambiguous_but_different_path(self):
        self.AnotherClass.__module__ += ".foo"
        self.registry.add_class(self.Class)
        self.registry.add_class(self.SubClass)
        self.registry.add_class(self.AnotherClass)
        prop1 = self.registry.get("properties.Class.prop1")
        prop2 = self.registry.get("properties.Class.prop2")
        self.assertTrue(prop1 is self.Class.prop1)
        self.assertTrue(prop2 is self.Class.prop2)
        prop1 = self.registry.get("SubClass.prop1")
        prop2 = self.registry.get("SubClass.prop2")
        self.assertTrue(prop1 is self.SubClass.prop1)
        self.assertTrue(prop2 is self.SubClass.prop2)
        prop1 = self.registry.get("foo.Class.prop1")
        prop2 = self.registry.get("foo.Class.prop2")
        self.assertTrue(prop1 is self.AnotherClass.prop1)
        self.assertTrue(prop2 is self.AnotherClass.prop2)

    def test_get_ambiguous_but_different_path_with_namespace(self):
        self.AnotherClass.__module__ += ".foo"
        self.registry.add_class(self.Class)
        self.registry.add_class(self.SubClass)
        self.registry.add_class(self.AnotherClass)
        prop1 = self.registry.get("Class.prop1", "storm.tests.properties")
        prop2 = self.registry.get("Class.prop2", "storm.tests.properties.bar")
        self.assertTrue(prop1 is self.Class.prop1)
        self.assertTrue(prop2 is self.Class.prop2)
        prop1 = self.registry.get("Class.prop1", "storm.tests.properties.foo")
        prop2 = self.registry.get(
            "Class.prop2", "storm.tests.properties.foo.bar")
        self.assertTrue(prop1 is self.AnotherClass.prop1)
        self.assertTrue(prop2 is self.AnotherClass.prop2)

    def test_class_is_collectable(self):
        self.AnotherClass.__module__ += ".foo"
        self.registry.add_class(self.Class)
        self.registry.add_class(self.AnotherClass)
        del self.AnotherClass
        gc.collect()
        prop1 = self.registry.get("prop1")
        prop2 = self.registry.get("prop2")
        self.assertTrue(prop1 is self.Class.prop1)
        self.assertTrue(prop2 is self.Class.prop2)

    def test_add_property(self):
        self.registry.add_property(self.Class, self.Class.prop1, "custom_name")
        prop1 = self.registry.get("Class.custom_name")
        self.assertEqual(prop1, self.Class.prop1)
        self.assertRaises(PropertyPathError, self.registry.get, "Class.prop1")


class PropertyPublisherMetaTest(TestHelper):

    def setUp(self):
        TestHelper.setUp(self)

        class Base(metaclass=PropertyPublisherMeta):
            pass

        class Class(Base):
            __storm_table__ = "mytable"
            prop1 = Property("column1", primary=True)
            prop2 = Property()

        class SubClass(Class):
            __storm_table__ = "mysubtable"

        self.Class = Class
        self.SubClass = SubClass

        class Class(Class):
            __module__ += ".foo"
            prop3 = Property("column3")

        self.AnotherClass = Class

        self.registry = Base._storm_property_registry

    def test_get_empty(self):
        self.assertRaises(PropertyPathError, self.registry.get, "unexistent")

    def test_get_subclass(self):
        prop1 = self.registry.get("SubClass.prop1")
        prop2 = self.registry.get("SubClass.prop2")
        self.assertTrue(prop1 is self.SubClass.prop1)
        self.assertTrue(prop2 is self.SubClass.prop2)

    def test_get_ambiguous(self):
        self.assertRaises(PropertyPathError, self.registry.get, "Class.prop1")
        self.assertRaises(PropertyPathError, self.registry.get, "Class.prop2")



import queue
from unittest import TestCase
import threading
import time

from storm.wsgi import make_app

class TestMakeApp(TestCase):

    def stub_app(self, environ, start_response):
        if getattr(self, 'in_request', None):
            self.in_request()
        getattr(self, 'calls', []).append('stub_app')
        start_response('200 OK', [])
        yield ''
        if getattr(self, 'in_generator', None):
            self.in_generator()

    def stub_start_response(self, status, headers):
        pass

    def test_find_timeline_outside_request(self):
        app, find_timeline = make_app(self.stub_app)
        # outside a request, find_timeline returns nothing:
        self.assertEqual(None, find_timeline())

    def test_find_timeline_in_request_not_set(self):
        # In a request, with no timeline object in the environ, find_timeline
        # returns None:
        app, find_timeline = make_app(self.stub_app)
        self.in_request = lambda:self.assertEqual(None, find_timeline())
        self.calls = []
        list(app({}, self.stub_start_response))
        # And we definitely got into the call:
        self.assertEqual(['stub_app'], self.calls)

    def test_find_timeline_set_in_environ(self):
        # If a timeline object is known, find_timeline finds it:
        app, find_timeline = make_app(self.stub_app)
        timeline = FakeTimeline()
        self.in_request = lambda:self.assertEqual(timeline, find_timeline())
        list(app({'timeline.timeline': timeline}, self.stub_start_response))

    def test_find_timeline_set_in_environ_during_generator(self):
        # If a timeline object is known, find_timeline finds it:
        app, find_timeline = make_app(self.stub_app)
        timeline = FakeTimeline()
        self.in_generator = lambda:self.assertEqual(timeline, find_timeline())
        list(app({'timeline.timeline': timeline}, self.stub_start_response))

    def test_timeline_is_replaced_in_subsequent_request(self):
        app, find_timeline = make_app(self.stub_app)
        timeline = FakeTimeline()
        self.in_request = lambda:self.assertEqual(timeline, find_timeline())
        list(app({'timeline.timeline': timeline}, self.stub_start_response))

        # Having left the request, the timeline is left behind...
        self.assertEqual(timeline, find_timeline())
        # ... but only until the next request comes through.
        timeline2 = FakeTimeline()
        self.in_request = lambda:self.assertEqual(timeline2, find_timeline())
        list(app({'timeline.timeline': timeline2}, self.stub_start_response))

    def test_lookups_are_threaded(self):
        # with two threads in a request at once, each only sees their own
        # timeline.
        app, find_timeline = make_app(self.stub_app)
        errors = queue.Queue()
        sync = threading.Condition()
        waiting = []
        def check_timeline():
            timeline = FakeTimeline()
            def start_response(status, headers):
                # Block on the condition, so all test threads are in
                # start_response when the test resumes.
                sync.acquire()
                waiting.append('x')
                sync.wait()
                sync.release()
                found_timeline = find_timeline()
                if found_timeline != timeline:
                    errors.put((found_timeline, timeline))
            list(app({'timeline.timeline': timeline}, start_response))
        t1 = threading.Thread(target=check_timeline)
        t2 = threading.Thread(target=check_timeline)
        t1.start()
        try:
            t2.start()
            try:
                while True:
                    sync.acquire()
                    if len(waiting) == 2:
                        break
                    sync.release()
                    time.sleep(0)
                sync.notify()
                sync.notify()
                sync.release()
            finally:
                t2.join()
        finally:
            t1.join()
        if errors.qsize():
            found_timeline, timeline = errors.get(False)
            self.assertEqual(timeline, found_timeline)


class FakeTimeline:
    """A fake Timeline.

    We need this because we can't use plain object instances as they can't be
    weakreferenced.
    """



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import weakref

from storm import has_cextensions


__all__ = ["EventSystem"]


class EventSystem:
    """A system for managing hooks that are called when events are emitted.

    Hooks are callables that take the event system C{owner} as their first
    argument, followed by the arguments passed when emitting the event,
    followed by any additional C{data} arguments given when registering the
    hook.

    Hooks registered for a given event C{name} are stored without ordering:
    no particular call order may be assumed when an event is emitted.
    """

    def __init__(self, owner):
        """
        @param owner: The object that owns this event system.  It is passed
            as the first argument to each hook function.
        """
        self._owner_ref = weakref.ref(owner)
        self._hooks = {}

    def hook(self, name, callback, *data):
        """Register a hook.

        @param name: The name of the event for which this hook should be
            called.
        @param callback: A callable which should be called when the event is
            emitted.
        @param data: Additional arguments to pass to the callable, after the
            C{owner} and any arguments passed when emitting the event.
        """
        callbacks = self._hooks.get(name)
        if callbacks is None:
            self._hooks.setdefault(name, set()).add((callback, data))
        else:
            callbacks.add((callback, data))

    def unhook(self, name, callback, *data):
        """Unregister a hook.

        This ignores attempts to unregister hooks that were not already
        registered.

        @param name: The name of the event for which this hook should no
            longer be called.
        @param callback: The callable to unregister.
        @param data: Additional arguments that were passed when registering
            the callable.
        """
        callbacks = self._hooks.get(name)
        if callbacks is not None:
            callbacks.discard((callback, data))

    def emit(self, name, *args):
        """Emit an event, calling any registered hooks.

        @param name: The name of the event.
        @param args: Additional arguments to pass to hooks.
        """
        owner = self._owner_ref()
        if owner is not None:
            callbacks = self._hooks.get(name)
            if callbacks:
                for callback, data in tuple(callbacks):
                    if callback(owner, *(args+data)) is False:
                        callbacks.discard((callback, data))


if has_cextensions:
    from storm.cextensions import EventSystem
